from typing import List, Dict, Callable, Iterator

from collections import defaultdict

import warnings

import time

import os

import copy

import hashlib

import datetime

from zoneinfo import ZoneInfo

import colorama

import uuid

import sentry_sdk

import tzdata

from nsj_gcf_utils.json_util import json_loads, json_dumps, JsonLoadException

from nsj_rest_lib.service.service_base import ServiceBase
from nsj_rest_lib.dto.dto_base import DTOBase
from nsj_rest_lib.descriptor.dto_field import DTOField
from nsj_rest_lib.descriptor.dto_list_field import DTOListField
from nsj_rest_lib.exception import AfterRecordNotFoundException

from sqlalchemy.exc import ProgrammingError

from nsj_integracao_api_client.infra.telemetria import (
    TelemetriaService, obter_telemetria, inicializar_telemetria
)

from nsj_integracao_api_client.service.integrador_cfg import (
    _entidades_filtros_integracao,  _entidades_integracao,
      _entidades_particionadas_por_empresa,
    _entidades_particionadas_por_estabelecimento, _entidades_particionadas_por_grupo,
    _cfg_filtros_to_dto_filtro, _ignorar_integridade, _entidades_blob,
    medir_tempo, Environment, TAMANHO_PAGINA, _E_CHECK_INT,
    TipoVerificacaoIntegridade
)

from nsj_integracao_api_client.service.cfg.tribo_pessoas import (
    _entidades_meurh_trab, _entidades_ponto, _entidades_avaliacao, _entidades_portal
)
from nsj_integracao_api_client.service.cfg.tribo_materiais import _entidades_tribo_materiais
from nsj_integracao_api_client.service.cfg.tribo_servicos import (
    _entidades_tribo_servicos, _entidades_atendimento
)

from nsj_integracao_api_client.infra.api_client import ApiClient

from nsj_integracao_api_client.infra.token_service import TokenService

from nsj_integracao_api_client.infra.injector_factory import InjectorFactory

from nsj_integracao_api_client.dao.integracao import IntegracaoDAO

from nsj_integracao_api_client.infra.debug_utils import DebugUtils as _du


colorama.init(autoreset=True)

out_func: Callable = print

_is_console = False


class IntegradorService():

    _injector : InjectorFactory = None

    _dao_intg: IntegracaoDAO = None

    _tz_br: ZoneInfo = None

    _token_service: TokenService = None

    _api_client: ApiClient = None
    
    _web_data_service = None  # Adapter para acesso direto ao banco Web

    _api_key: str = None

    _tenant: int = None

    # Caso se deseje usar um tenant diferente do especificado no token
    _forced_tenant: int = None

    _filtros_particionamento: list = None

    _save_point: dict = {}

    _ignored_fields : list = ["tenant", "lastupdate"]

    _detalhar_diferencas: bool

    _trace: bool

    _interromper_execucao: bool

    _em_execucao: bool

    _menos_log: bool

    _adicionar_data_log: bool

    _env: Environment

    _telemetria: TelemetriaService = None

    _cache_entidades_integracao : List[str] = []

    def __init__(
        self,
        injector: InjectorFactory,
        log,
        env: Environment = Environment.PROD,
        forced_tenant: int = None,
        menos_log: bool = False,
        adicionar_data_log: bool = False
    ):
        self._injector = injector
        env_cfg = self._integracao_dao().recuperar_configuracao_ambiente()
        self._env = Environment(env_cfg) if env_cfg else env
        self._logger = log
        self._forced_tenant = forced_tenant
        self._tz_br = ZoneInfo("America/Sao_Paulo")
        self._detalhar_diferencas = False
        self._trace = False
        self._api_client = ApiClient(env)
        self._token_service = TokenService()
        self._interromper_execucao = False
        self._em_execucao = False
        self._menos_log = menos_log
        self._adicionar_data_log = adicionar_data_log
        self.atualizar_entidades_integracao()

        if env == Environment.PROD:

            emp = {}
            try:
                emp = self._integracao_dao().recuperar_dados_empresa_licenciamento()
            except Exception:
                pass

            inicializar_telemetria(
                url_api="http://telemetria.nasajon.com.br/api/events",
                empresadetentora=f"{emp['Codigo']}-{emp['Nome']}" if emp else "Indefinida",
                cnpjdetentora=emp.get("Cnpj","Indefinido")
            )
            self._telemetria = obter_telemetria()


    def _log(self, msg):
        if self._adicionar_data_log and msg.replace("\n", "").replace("\r", "") != "":
            msg = f"[{datetime.datetime.now(self._tz_br).strftime('%Y-%m-%d %H:%M:%S')}] {msg}"
        self._logger.mensagem(msg)


    def _processa_mensagens(self, msgs: Iterator[str]):
        for msg in msgs:
            if not self._menos_log:
                self._log(msg)


    def _is_missing_table_error(self, ex: ProgrammingError) -> bool:
        detail = None

        if getattr(ex, 'orig', None) is not None and getattr(ex.orig, 'args', None):
            detail = ex.orig.args[0]
        elif getattr(ex, 'args', None):
            detail = ex.args[0]

        if isinstance(detail, dict):
            return detail.get('C') == '42P01'

        return False


    def _carregar_savepoint(self):
        try:
            with open('savepoint.json', 'r', encoding='utf-8') as f:
                self._save_point = json_loads(f.read())
                entidade_salva = list(self._save_point.keys())[0]
                self._log(f"Savepoint carregado para : {entidade_salva }")
        except FileNotFoundError:
            self._save_point = {}


    def _trace_check(self, filename, content):

        _du.conditional_trace(
            condition=_E_CHECK_INT or self._trace,
            func=_du.save_to_file,
            filename=filename,
            content=content
        )


    def _detail_check(self, filename, content):

        _du.conditional_trace(
            condition=_E_CHECK_INT or self._detalhar_diferencas,
            func=_du.save_to_file,
            filename=filename,
            content=content
        )


    def _entidades_sistema(self, cod_sistema: str)-> list:

        match cod_sistema:
            case 'meurh' | 'meutrabalho': return _entidades_meurh_trab
            case 'ponto': return _entidades_ponto
            case 'portal': return _entidades_portal
            case 'avaliacao': return _entidades_avaliacao
            case 'pedidos' | 'estoqueapi': return _entidades_tribo_materiais
            case 'servicos' | 'crm' | 'gp': return _entidades_tribo_servicos
            case 'atendimento': return _entidades_atendimento
            case _:  return []


    def _carrega_entidades_sistemas_contratados(self)-> list:
        """
        Carrega as entidades disponíveis para integração de acordo com os sistemas contratados pelo tenant.
        Consulta a API para obter os sistemas contratados e, para cada sistema, obtém a lista de entidades
        correspondentes. Retorna uma lista única de entidades que devem ser integradas.
        """

        ambiente = self._api_client.sistemas_contratados_ambiente(self.api_key, self.tenant)

        _entidades = set()
        if 'sistemas' in ambiente:
            for sistema in ambiente['sistemas']:
                _entidades |= set(self._entidades_sistema(sistema))

        return list(_entidades)


    def _entidades_integracao(self)-> list:
        """
        Retorna a lista de entidades para integrar de acordo com a ordenação lógica das entidades configuradas
        """

        if not self._cache_entidades_integracao:

            _entidades_contratacoes = self._carrega_entidades_sistemas_contratados()
            self._cache_entidades_integracao = [_entidade for _entidade in _entidades_integracao if _entidade in _entidades_contratacoes]

        return self._cache_entidades_integracao


    def _fields_to_load_erp(self, service: ServiceBase):

        _campos_erp = set([
            _field['column_name'] for _field in self._integracao_dao().listar_campos_entidade(service._entity_class.table_name)
        ])

        _root_fields = sorted(set(service._dto_class.fields_map.keys()) & _campos_erp)

        fields = {}
        fields.setdefault("root",set(_root_fields))

        for _related_entity, _related_list_fields in service._dto_class.list_fields_map.items():
            fields["root"].add(_related_entity)
            fields.setdefault(_related_entity, set())

            _campos_erp_related = set([
                _field['column_name'] for _field in self._integracao_dao().listar_campos_entidade(_related_list_fields.entity_type.table_name)
            ])
            _related_fields = sorted(set(_related_list_fields.dto_type.fields_map.keys()) & _campos_erp_related)
            for _related_field in _related_fields:
                fields["root"].add(f"{_related_entity}.{_related_field}")
                fields[_related_entity].add(_related_field)

        return fields


    def _fields_to_load(self, dto_class: DTOBase) -> dict:
        warnings.warn("_fields_to_load está depreciado e será removido em breve.", DeprecationWarning)

        fields = {}
        fields.setdefault("root", set(dto_class.fields_map.keys()))

        for _related_entity, _related_list_fields in dto_class.list_fields_map.items():
            fields["root"].add(_related_entity)
            fields.setdefault(_related_entity, set())
            _related_fields = _related_list_fields.dto_type.fields_map.keys()
            for _related_field in _related_fields:
                fields["root"].add(f"{_related_entity}.{_related_field}")
                fields[_related_entity].add(_related_field)

        return fields


    def _integracao_dao(self):
        if self._dao_intg is None:
            self._dao_intg = self._injector.integracao_dao()
        return self._dao_intg


    @property
    def api_key(self):

        if self._api_key is None:
            self._api_key = self._integracao_dao().recuperar_token()

        return self._api_key


    @property
    def tenant(self):

        if self._forced_tenant is not None:
            return self._forced_tenant

        if self._tenant is None:
            decoded_token = self._token_service.decode_token(self.api_key)
            self._tenant = decoded_token["tenant_id"]

        return self._tenant


    def _integracao_foi_configurada(self):
        return self._integracao_dao().integracao_configurada()


    def _validar_grupos_empresariais(self, grupos) -> List[Dict[str, str]]:

        grupos_cadastrados = self._integracao_dao().listar_grupos_empresariais(grupos)
        _cods = [grupo['codigo'] for grupo in grupos_cadastrados]
        _grupos_faltantes = [grupo for grupo in grupos if grupo not in _cods]
        assert len(_grupos_faltantes)==0, f"Grupo(s) '{','.join(_grupos_faltantes)}' não encontrado(s)."
        return grupos_cadastrados
    
    def atualizar_entidades_integracao(self):
        """Atualiza a lista de entidades para integração."""
        
        try:
            
            entidades_registrar = self._entidades_integracao()
            entidades_registradas_list = self._integracao_dao().listar_entidades_integracao()
            # Extrai o nome se vier como dict
            entidades_registradas = []
            
            for e in entidades_registradas_list:
                
                if isinstance(e, dict):
                    nome = e.get('nome') or e.get('entidade') or str(e)
                    entidades_registradas.append(nome)
                else:
                    entidades_registradas.append(str(e))

            # Normaliza para comparação
            entidades_registradas_norm = set(e.strip().lower() for e in entidades_registradas)

            for entidade in entidades_registrar:
                if entidade.strip().lower() not in entidades_registradas_norm:
                    self._log(f"Entidade '{entidade}' não está registrada no banco!")
                    self._integracao_dao().registra_entidade_integracao(entidade)
                    self._log(f"Entidade '{entidade}' registrada com sucesso.")

        except Exception as e:
            self._log(f"Erro ao atualizar entidades de integração: {str(e)}")
            sentry_sdk.capture_exception(e)

    def registrar_entidades_integracao(self):
        for entidade in _entidades_integracao:

            self._integracao_dao().registra_entidade_integracao(entidade)

            _dto = self._injector.dto_for(entidade, False)

            for field in _dto.list_fields_map.values():
                _sub_entity = field.entity_type.table_name
                self._integracao_dao().registra_entidade_integracao(_sub_entity)


    def executar_instalacao(self, chave_ativacao: str, grupos: List[str]):

        assert chave_ativacao, "Chave de ativação não pode ser vazia."
        self._log(f"Executando instalação com a chave de ativação: {chave_ativacao}")

        assert not self._integracao_foi_configurada(), "Integração já instalada anteriormente."
        
        # Valida grupos empresariais localmente antes de chamar a API
        if grupos:
            grupos_cadastrados = self._validar_grupos_empresariais(grupos)
            grupos_para_ativar = grupos  # Usa os códigos/UUIDs originais
        else:
            grupos_cadastrados = self._integracao_dao().listar_grupos_empresariais()
            grupos_para_ativar = [grupo['codigo'] for grupo in grupos_cadastrados]
        
        # Primeiro gera o token para obter o tenant_id
        _token: str = self._api_client.gerar_token_tenant(chave_ativacao)
        decoded_token = self._token_service.decode_token(_token)
        tenant_id = decoded_token['tenant_id']
        
        self._log(f"Token gerado para tenant {tenant_id}")

        if self._web_data_service:
            # Modo servidor: chama a API de ativação
            try:
                self._log(f"Ativando integração via API para tenant {tenant_id}...")
                resultado = self._api_client.ativar_integracao(
                    tenant=tenant_id,
                    chave_ativacao=chave_ativacao,
                    grupos_empresariais=grupos_para_ativar,
                    api_key=_token
                )
                
                self._log(f"Instalação efetuada com sucesso para o tenant '{tenant_id}'.")
                self._log(f"Grupos ativados: {resultado.get('grupos_ativados', 0)}")
                self._log(f"Entidades registradas: {resultado.get('entidades_registradas', 0)}")
                
                self.atualizar_entidades_integracao()
                
            except Exception as e:
                self._log(f"Erro ao ativar integração: {str(e)}")
                raise
        else:
            # Modo utilitário (desktop/JobManager): operações diretas no banco
            try:
                self._integracao_dao().begin()

                _ids = [grupo['id'] for grupo in grupos_cadastrados]
                self._integracao_dao().registrar_grupos_empresariais(_ids)
                self._log(f"Grupos empresariais registrados: {len(_ids)}")

                self._integracao_dao().registra_token_tenant(_token)
                self._log("Token de integração registrado.")

                self._integracao_dao().commit()
            except Exception as e:
                self._integracao_dao().rollback()
                self._log(f"Erro ao ativar integração: {str(e)}")
                raise

            self.registrar_entidades_integracao()
            self._log(f"Entidades de integração registradas.")

            self.atualizar_entidades_integracao()
            self._log(f"Instalação efetuada com sucesso para o tenant '{tenant_id}'.")


    def ativar_grupos_empresariais(self, grupos: List[str]):

        assert self._integracao_foi_configurada(), "Integração não configurada!"

        # Valida grupos empresariais localmente
        if grupos:
            grupos_cadastrados = self._validar_grupos_empresariais(grupos)
            grupos_para_ativar = grupos  # Usa os códigos/UUIDs originais
        else:
            grupos_cadastrados = self._integracao_dao().listar_grupos_empresariais()
            grupos_para_ativar = [grupo['codigo'] for grupo in grupos_cadastrados]

        if self._web_data_service:
            # Modo servidor: usa a API
            self._log(f"Ativando grupos empresariais via API: {grupos_para_ativar}")
            resultado = self._api_client.ativar_grupos_empresariais(
                tenant=self._tenant,
                grupos=grupos_para_ativar,
                api_key=self._api_key
            )
            
            self._log(f"Grupos empresariais ativados via API: '{','.join(grupos_para_ativar)}'.")
            self._log(f"Total ativados: {resultado.get('total_ativados', 0)}")
        else:
            # Modo utilitário (desktop/JobManager): acesso direto ao banco
            _ids = [grupo['id'] for grupo in grupos_cadastrados]
            self._integracao_dao().registrar_grupos_empresariais(_ids)
            
            if grupos:
                self._log(f"Grupos empresariais ativados (modo local): '{','.join(grupos)}'.")
            else:
                _codigos = [grupo['codigo'] for grupo in grupos_cadastrados]
                self._log(f"Grupos empresariais ativados (modo local): '{','.join(_codigos)}'.")


    def desativar_grupos_empresariais(self, grupos: List[str]):

        assert self._integracao_foi_configurada(), "Integração não configurada!"
        assert grupos, "Grupos não podem ser vazios!"

        # Valida grupos empresariais localmente
        grupos_cadastrados = self._validar_grupos_empresariais(grupos)
        grupos_para_desativar = grupos  # Usa os códigos/UUIDs originais

        if self._web_data_service:
            # Modo servidor: usa a API
            self._log(f"Desativando grupos empresariais via API: {grupos_para_desativar}")
            resultado = self._api_client.desativar_grupos_empresariais(
                tenant=self._tenant,
                grupos=grupos_para_desativar,
                api_key=self._api_key
            )
            
            self._log(f"Grupos empresariais desativados via API: '{','.join(grupos)}'.")
            self._log(f"Total desativados: {resultado.get('total_desativados', 0)}")
        else:
            # Modo utilitário (desktop/JobManager): acesso direto ao banco
            _ids = [grupo['id'] for grupo in grupos_cadastrados]
            self._integracao_dao().desativar_grupos_empresariais(_ids)
            
            self._log(f"Grupos empresariais desativados (modo local): '{','.join(grupos)}'.")


    def desativar_integracao(self):
        """
        Desativa a integração removendo o token de acesso.
        """
        assert self._integracao_foi_configurada(), "Integração não configurada!"

        if self._web_data_service:
            # Modo servidor: usa a API
            self._log("Desativando integração via API...")
            self._api_client.desativar_integracao(
                tenant=self._tenant,
                api_key=self._api_key
            )
            self._log("Integração desativada com sucesso via API.")
        else:
            # Modo utilitário (desktop/JobManager): acesso direto ao banco
            try:
                self._integracao_dao().begin()
                self._integracao_dao().remove_token_tenant()
                self._integracao_dao().commit()
                
                self._log("Integração desativada com sucesso (modo local).")
                
            except Exception as db_error:
                self._integracao_dao().rollback()
                self._log(f"Erro ao desativar integração via banco: {str(db_error)}")
                raise


    def _filtro_particionamento_de(self, entidade: str):

        if self._filtros_particionamento is None:
            _dados_part = self._integracao_dao().listar_dados_particionamento()

            assert _dados_part, "Não existem entidades empresariais cadastradas para integração!"

            self._filtros_particionamento = [
                {'grupoempresarial' : ",".join(list(map(lambda i: str(i["grupoempresarial"]), _dados_part)))},
                {'empresa' : ",".join(list(map(lambda i: str(i["empresa"]), _dados_part)))},
                {'estabelecimento' : ",".join(list(map(lambda i: str(i["estabelecimento"]), _dados_part)))}
            ]

        if entidade in _entidades_particionadas_por_grupo:
            return  self._filtros_particionamento[0]

        if entidade in _entidades_particionadas_por_empresa:
            return self._filtros_particionamento[1]

        if entidade in _entidades_particionadas_por_estabelecimento:
            return self._filtros_particionamento[2]

        return {}


    def _filtros_integracao_cfg(self, entidade: str):

        _filtro_salvo = self._integracao_dao().filtros_integracao_entidade(entidade)

        if not isinstance(_filtro_salvo, list):
            self._log(f"O filtro salvo para a entidade '{entidade}' deve ser uma lista, {_filtro_salvo} fornecido.")
            _filtro_salvo = []

        _filtros = (
            _entidades_filtros_integracao[entidade] +
            _filtro_salvo
        )
        #Garante unicidade do campo no filtro
        dict_filtros = {}
        for _item in _filtros:
            if "campo" in _item and "valor" in _item and "operador" in _item:
                dict_filtros[_item['campo']] = copy.copy(_item)
            else:
                self._log(f"O filtro salvo para a entidade '{entidade}' deve conter campo,valor e operador': {_item} fornecido.")

        return list(dict_filtros.values())


    def _dto_to_api(
        self,
        campos: Dict[str, List[str]],
        data: List[DTOBase]
    ) -> List[dict]:
        # Converte os objetos DTO para dicionários e adiciona o tenant
        transformed_data = []
        for dto in data:
            dto.tenant = self.tenant
            dto_dict = dto.convert_to_dict(campos)

            # Implementado devido a dualidade de conversão de campos json na API que podem ser convertidos com str/dict

            if "created_by" in dto_dict and not dto_dict["created_by"] is None:
                if not isinstance(dto_dict["created_by"], dict):
                    try:
                        _value_dict = json_loads(dto_dict["created_by"])
                        dto_dict["created_by"] = _value_dict
                    except (TypeError, ValueError, JsonLoadException):
                        dto_dict["created_by"] = {"id": dto_dict["created_by"]}

            if "updated_by" in dto_dict and not dto_dict["updated_by"] is None:
                if not isinstance(dto_dict["updated_by"], dict):
                    try:
                        _value_dict = json_loads(dto_dict["updated_by"])
                        dto_dict["updated_by"] = _value_dict
                    except (TypeError, ValueError, JsonLoadException):
                        dto_dict["updated_by"] = {"id": dto_dict["updated_by"]}
            
            # Adicionando o tenant pois agora o restlib não considera a coluna no ambiente desktop.       
            if 'tenant' not in dto_dict:
                dto_dict['tenant'] = self.tenant
                
            transformed_data.append(dto_dict)

        return transformed_data


    def _save_point_for(self, tabela: str):
        return self._save_point.get(tabela, None)


    def _do_save_point(self, tabela: str, chave):
        self._save_point[tabela] = chave
        with open('savepoint.json', 'w', encoding='utf-8') as f:
            f.write(f'{{ "{tabela}": "{chave}" }} ' if chave else f'{{ "{tabela}": null }} ')


    def _save_point_clear(self):
        self._save_point.clear()
        if os.path.exists('savepoint.json'):
            os.remove('savepoint.json')


    def interromper_execucao(self):
        self._interromper_execucao = True
        self._em_execucao = False


    def em_execucao(self):
        return self._em_execucao


    def _atualiza_ultima_integracao(self, entidade: str, filtros : list):
        self._integracao_dao().atualiza_ultima_integracao(entidade, filtros)

        _dto = self._injector.dto_for(entidade, False)

        for field in _dto.list_fields_map.values():
            _sub_entity = field.entity_type.table_name
            self._integracao_dao().atualiza_ultima_integracao(_sub_entity, [])


    def _atualiza_data_ultima_integracao(self, entidade: str, data: datetime.datetime, filtros: list):
        self._integracao_dao().atualiza_data_ultima_integracao(entidade, data, filtros)

        _dto = self._injector.dto_for(entidade, False)

        for field in _dto.list_fields_map.values():
            _sub_entity = field.entity_type.table_name
            self._integracao_dao().atualiza_data_ultima_integracao(_sub_entity, data, [])


    def _enviar_dados(self, dict_data: list, entidade: str, parar_caso_erros: bool = False) -> list:
        _erros = []
        try:
            # Usa WebDataService se disponível (acesso direto ao banco Web), senão usa ApiClient (HTTP)
            if self._web_data_service:
                self._web_data_service.upsert_dados(entidade=entidade, dados=dict_data)
            else:
                self._api_client.enviar_dados(dict_data, entidade, self.api_key)
        except Exception:
            # Se o envio em bloco falhar, tenta o envio item por item
            for _item in dict_data:
                try:
                    if self._web_data_service:
                        self._web_data_service.upsert_dados(entidade=entidade, dados=[_item])
                    else:
                        self._api_client.enviar_dados([_item], entidade, self.api_key)
                except Exception as e:
                    if parar_caso_erros:
                        raise e
                    else:
                        self._log("")
                        self._log(f"\n{'-'*80}\nErro:\n{str(e)}\n{'-'*80 if _item == dict_data[-1] else ''}")
                        self._log("")
                        _erros.append(e)
                        sentry_sdk.capture_exception(e)
        return _erros


    def _apagar_dados_bulk(self, dict_data: list, entidade: str, parar_caso_erros: bool = False):
        _erros = []
        try:
            # Tenta o envio em bloco
             self._api_client.apagar_dados_bulk(dict_data, entidade, self.api_key, self.tenant)
        except Exception as e:
            if parar_caso_erros:
                raise e
            else:
                self._log("")
                self._log(f"\n{'-'*80}\nErro:\n{str(e)}\n{'-'*80}")
                self._log("")
                _erros.append(e)
                sentry_sdk.capture_exception(e)


    def _apagar_dados(self, dict_data: list, entidade: str, parar_caso_erros: bool = False) -> list:
        _erros = []
        try:
            # Tenta o envio em bloco
            mensagens = self._api_client.apagar_dados(dict_data, entidade, self.api_key, self.tenant)
            self._processa_mensagens(mensagens)
        except Exception:
            # Se o envio em bloco falhar, tenta o envio item por item
            for _item in dict_data:
                try:
                    mensagens = self._api_client.apagar_dados([_item], entidade, self.api_key, self.tenant)
                    self._processa_mensagens(mensagens)
                except Exception as e:
                    if parar_caso_erros:
                        raise e
                    else:
                        self._log("")
                        self._log(f"\n{'-'*80}\nErro:\n{str(e)}\n{'-'*80 if _item == dict_data[-1] else ''}")
                        self._log("")
                        _erros.append(e)
                        sentry_sdk.capture_exception(e)

        return _erros


    def _enviar_blobs(self, ids: list, files: list ,entidade: str, campo: str, parar_caso_erros: bool = False):
        _erros = []
        try:
            # Tenta o envio em bloco
            self._api_client.enviar_blobs(ids, files ,entidade, campo, self.tenant, self.api_key)
        except Exception:
            # Se o envio em bloco falhar, tenta o envio item por item
            for i, file in enumerate(files):
                try:
                    self._api_client.enviar_blobs([ids[i]], [file] ,entidade, campo, self.tenant, self.api_key)
                except Exception as e:
                    if parar_caso_erros:
                        raise e
                    else:
                        self._log("")
                        self._log(f"\n{'-'*80}\nErro:\n{str(e)}\n{'-'*80}")
                        self._log("")
                        _erros.append(e)
                        sentry_sdk.capture_exception(e)

        return _erros


    def _processar_blobs(self, entidade: str, service, dict_data: list, avaliar_diferencas: bool = True, parar_caso_erros: bool = False):

        if entidade in _entidades_blob:

            if service is None:
                service = self._injector.service_for(entidade, True)

            _campos = _entidades_blob[entidade]
            _ids = [i['id'] for i in dict_data]
            _pk_banco = service._entity_class.pk_field

            self._log(f"Verificando se houve alterações nos blobs da entidade {entidade}.")

            for _campo in _campos:

                _ids_blobs = []
                _file_blobs = []

                if avaliar_diferencas:
                    _ids_to_send_blob = self._api_client.consultar_hash_blob(_ids, entidade, _campo, self.tenant, self.api_key)

                    _hashes_blob_remoto = _ids_to_send_blob['result']

                    _hashes_blob_local = self._integracao_dao().listar_blobs_entidade(_pk_banco, _campo, entidade, _ids)

                    # comparar md5
                    for blob_local in _hashes_blob_local:
                        id_local = blob_local['id']
                        hash_local = blob_local['hash']
                        # Procura o hash remoto correspondente
                        hash_remoto = next((item['hash'] for item in _hashes_blob_remoto if item['id'] == id_local), None)
                        if hash_remoto != hash_local:
                            _conteudo =   blob_local['blob'] if blob_local['blob'] is not None else b""
                            _ids_blobs.append(blob_local['id'])
                            _file_blobs.append(('files', (blob_local['id'], _conteudo, 'application/octet-stream')))
                else:
                    _hashes_blob_local = self._integracao_dao().listar_blobs_entidade(_pk_banco, _campo, entidade, _ids)
                    for blob_local in _hashes_blob_local:
                        _conteudo =   blob_local['blob'] if blob_local['blob'] is not None else b""
                        _ids_blobs.append(blob_local['id'])
                        _file_blobs.append(('files', (blob_local['id'], _conteudo, 'application/octet-stream')))


                self._log(f"Enviando blobs do campo {_campo} para a api.")
                if _ids_blobs:
                    self._enviar_blobs(
                        _ids_blobs,
                        _file_blobs,
                        entidade,
                        _campo,
                        parar_caso_erros
                    )


    def _telemetria_inicio_carga_inicial(
            self,
            correlation_id: str,
            qtd_entidades: int
        ):
        if self._telemetria:
            self._telemetria.evento_inicio_carga(
                correlation_id=correlation_id,
                entidades_processar=qtd_entidades,
                tenant=self.tenant,
                ambiente=self._env.value
            )


    def _telemetria_carga_entidade(
            self,
            correlation_id: str,
            entidade: str,
            ordem: int,
            total_atualizacoes: int
        ):
        # removido estava gerando excesso de métricas
        pass
        # if self._telemetria:
        #     self._telemetria.evento_carga_entidade(
        #         correlation_id=correlation_id,
        #         entidade=entidade,
        #         ordem_processamento=ordem,
        #         total_atualizacoes=total_atualizacoes
        #     )


    def _telemetria_fim_carga_inicial(
            self,
            correlation_id: str,
            qtd_entidades: int,
            qtd_registros: int,
            duracao_ms: int,
            status:str
        ):
        if self._telemetria:
            self._telemetria.evento_fim_carga(
                correlation_id=correlation_id,
                entidades_processadas=qtd_entidades,
                total_registros=qtd_registros,
                duracao_total_ms=duracao_ms,
                status=status
            )


    def _telemetria_inicio_integracao(
            self,
            correlation_id: str,
            qtd_entidades: int
        ):
        if self._telemetria:
            self._telemetria.evento_inicio_integracao(
                correlation_id=correlation_id,
                entidades_pendentes=qtd_entidades,
                tenant=self.tenant,
                ambiente=self._env.value
            )


    def _telemetria_integracao_entidade(
            self,
            correlation_id: str,
            entidade: str,
            ordem: int,
            total_exclusoes: int,
            total_atualizacoes: int,
            data_ultima_integracao: datetime.datetime,
        ):
        # removido estava gerando excesso de métricas
        pass
        # if self._telemetria:
        #     self._telemetria.evento_integracao_entidade(
        #         correlation_id=correlation_id,
        #         entidade=entidade,
        #         ordem_processamento=ordem,
        #         total_exclusoes=total_exclusoes,
        #         total_atualizacoes=total_atualizacoes,
        #         data_ultima_integracao=data_ultima_integracao
        #     )


    def _telemetria_fim_integracao(
            self,
            correlation_id: str,
            entidades_processadas: int,
            total_atualizacoes: int,
            total_exclusoes: int,
            duracao_ms: int,
            status:str
        ):
        if self._telemetria:
            self._telemetria.evento_fim_integracao(
                correlation_id=correlation_id,
                entidades_processadas=entidades_processadas,
                total_atualizacoes=total_atualizacoes,
                total_exclusoes=total_exclusoes,
                duracao_total_ms=duracao_ms,
                status=status
            )


    def _telemetria_inicio_verificacao_integridade(self):
        if self._telemetria:
            self._telemetria.evento_inicio_verificacao(
                correlation_id=None,
                entidades_verificar=15,
                tenant=self.tenant,
                ambiente=self._env.value,
                tipo_verificacao="HASH"
            )


    def _telemetria_fim_verificacao_integridade(self):
        if self._telemetria:
            self._telemetria.evento_fim_verificacao(
                correlation_id=None,
                entidades_verificadas=15,
                total_diferencas=30,
                duracao_total_ms=300000
            )


    def _status_execucao(self, erros: bool, _parar_caso_erro: bool, interrompido: bool):
        if interrompido:
            return "INTERROMPIDO"

        if not erros:
            return "FINALIZADO"

        if erros and _parar_caso_erro:
            return "FALHA"

        if erros and not _parar_caso_erro:
            return "FINALIZADO_COM_ERROS"


    @medir_tempo("Carga inicial")
    def executar_carga_inicial(self, entidades: list, parar_caso_erros: bool = False):
        self._interromper_execucao = False
        self._em_execucao = True

        inicio = time.perf_counter()
        _count_total = 0
        correlation_id = str(uuid.uuid4())

        _erros_envio = []
        entidades_carga_inicial = []

        try:
            assert self._integracao_foi_configurada(), "Integração não configurada!"

            _dao = self._integracao_dao()

            assert _dao.existem_grupos_empresariais_integracao_ativos(), "Nenhum grupo empresarial ativo para integração."
            assert _dao.listar_dados_particionamento(), "Entidades empresariais inválidas para integração."

            self._log(f"Executando carga inicial para o Tenant: {self.tenant} .")
            self._log(f"{len(self._entidades_integracao())} entidades para processar.")

            entidades_carga_inicial = copy.copy(self._entidades_integracao())

            # Remover entidades que nao devem ser processadas
            if entidades:
                for entidade in entidades:
                    assert entidade in self._entidades_integracao(), f"Entidade '{entidade}' não consta como entidade para integração!"

                for entidade in self._entidades_integracao():
                    if not entidade in entidades:
                        entidades_carga_inicial.remove(entidade)

            # Remover entidades que ja foram processadas
            if not entidades:
                self._carregar_savepoint()
                if self._save_point:
                    for entidade in self._entidades_integracao():
                        if not entidade in self._save_point:
                            entidades_carga_inicial.remove(entidade)
                        else:
                            break

            self._telemetria_inicio_carga_inicial(correlation_id, len(entidades_carga_inicial))

            for entidade in entidades_carga_inicial:

                # if not entidade in ['persona.adiantamentosavulsos','persona.trabalhadores']:
                #     continue

                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return

                _idx = self._entidades_integracao().index(entidade) + 1
                self._log(f"Efetuando carga {entidade}, {_idx} de {len(self._entidades_integracao())}.")
                _count = 0


                # Carregar dados paginados para integrar
                service = self._injector.service_for(entidade, True)
                fields = self._fields_to_load_erp(service)
                _filtros_particionamento = self._filtro_particionamento_de(entidade)
                _filtros_integracao_cfg = self._filtros_integracao_cfg(entidade)
                filters = _filtros_particionamento | _cfg_filtros_to_dto_filtro(_filtros_integracao_cfg)
                search_query = None

                #pagina = 0

                self._log("Extraindo dados para carga.")
                while True:

                    if self._interromper_execucao:
                        self._log("Processo interrompido pelo usuário.")
                        return

                    current_after = self._save_point_for(entidade)
                    try:
                        _data = service.list(
                                current_after,
                                TAMANHO_PAGINA,
                                fields,
                                None,
                                filters,
                                search_query=search_query,
                            )
                    except ProgrammingError as ex:
                        if self._is_missing_table_error(ex):
                            self._log(f"WARNING: {entidade}: tabela não encontrada no banco de dados. Pulando entidade.")
                            break
                        raise

                    _count = _count + len(_data)
                    _count_total = _count_total + _count

                    if len(_data)==0:
                        if current_after is None:
                            self._log("Sem dados para transferir, indo adiante.")
                        else:
                            self._log("Entidade integrada com sucesso.")
                        break

                    if self._interromper_execucao:
                        self._log("Processo interrompido pelo usuário.")
                        return

                    self._log(f"{_count} registro(s).")

                    dict_data = self._dto_to_api(fields, _data)

                    self._log("Enviando dados para a api.")
                    _erros_envio += self._enviar_dados(dict_data, entidade, parar_caso_erros)

                    self._processar_blobs(entidade, service, dict_data, False, parar_caso_erros)

                    # Aponta a leitura para a próxima página
                    _last = _data[-1]
                    self._do_save_point(entidade, getattr(_last, _last.pk_field))

                self._atualiza_ultima_integracao(entidade, _filtros_integracao_cfg)
                self._save_point_clear()
                self._telemetria_carga_entidade(
                    correlation_id,
                    entidade,
                    entidades_carga_inicial.index(entidade)+1,
                    _count
                )

            self._log(self._color("Carga inicial finalizada com sucesso!", "92", _is_console))

            if _erros_envio:
                self._log(self._color("Ocorreram erros que foram ignorados durante o processo, verifique:", 91, _is_console))
                self._log("")
                self._log("\n"+("-"*80)+"\n"+self._color("\n".join(str(e)+ "\n" + "-"*80 for e in _erros_envio), 91, _is_console))


        finally:
            fim = time.perf_counter()
            self._telemetria_fim_carga_inicial(
                correlation_id,
                len(entidades_carga_inicial),
                _count_total,
                (fim - inicio) * 1000,
                self._status_execucao(len(_erros_envio), parar_caso_erros, self._interromper_execucao)
            )
            self._em_execucao = False


    @medir_tempo("Integração")
    def executar_integracao_old(self):
        self._interromper_execucao = False
        self._em_execucao = True

        try:
            assert self._integracao_foi_configurada(), "Integração não configurada!"

            _dao = self._integracao_dao()

            assert _dao.existem_grupos_empresariais_integracao_ativos(), "Nenhum grupo empresarial ativo para integração."

            self._log(f"Executando integração para o Tenant: {self.tenant} .")

            entidades_pendentes_bd = _dao.listar_entidades_pendentes_integracao()

            # Não filtrar entidades filhas
            entidades_pendentes = {entidade: entidades_pendentes_bd[entidade] for entidade in _entidades_integracao if entidade in entidades_pendentes_bd.keys()}

            self._log(f"{len(entidades_pendentes)} entidades para processar." if entidades_pendentes else "Nenhuma entidade para processar.")
            _resumo = {}

            self._integracao_dir = f"integracao_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}"

            for entidade, data_ultima_integracao in entidades_pendentes.items():


                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return

                _idx = list(entidades_pendentes.keys()).index(entidade) + 1
                self._log(f"Integrando {entidade}, {_idx} de {len(entidades_pendentes)}.")
                _count = 0

                # Carregar dados paginados para integrar
                service = self._injector.service_for(entidade, True)
                current_after = None
                fields = self._fields_to_load_erp(service)
                _filtros_particionamento = self._filtro_particionamento_de(entidade)
                _filtros_integracao_cfg = self._filtros_integracao_cfg(entidade)
                filters = _filtros_particionamento | _cfg_filtros_to_dto_filtro(_filtros_integracao_cfg)
                search_query = None

                # Dados excluidos apos data_ultima_integracao
                _coluna_id = service._dto_class.fields_map[service._dto_class.pk_field].entity_field
                para_apagar = _dao.listar_dados_exclusao(_coluna_id, entidade, data_ultima_integracao)
                if para_apagar:
                    for i in range(0, len(para_apagar), TAMANHO_PAGINA):
                        bloco_para_apagar = para_apagar[i:i+TAMANHO_PAGINA]
                        _resumo[entidade] = _resumo.get(entidade, 0) + len(bloco_para_apagar)
                        mensagens = self._api_client.apagar_dados(bloco_para_apagar, entidade, self.api_key, self.tenant)
                        self._processa_mensagens(mensagens)

                # Dados alterados apos data_ultima_integracao
                filtro_atualizacao = filters.copy() if filters else {}
                filtro_atualizacao['lastupdate'] = data_ultima_integracao
                while True:

                    if self._interromper_execucao:
                        self._log("Processo interrompido pelo usuário.")
                        return

                    try:
                        _data = service.list(
                                current_after,
                                TAMANHO_PAGINA,
                                fields,
                                None,
                                filtro_atualizacao,
                                search_query=search_query,
                            )
                    except ProgrammingError as ex:
                        if self._is_missing_table_error(ex):
                            self._log(f"WARNING: {entidade}: tabela não encontrada no banco de dados. Pulando entidade.")
                            break
                        raise

                    _count = _count + len(_data)

                    if len(_data)==0:
                        break

                    self._log(f"{_count} registro(s).")
                    _resumo[entidade] = _resumo.get(entidade, 0) + _count

                    # Convertendo para o formato de dicionário (permitindo omitir campos do DTO) e add tenant
                    dict_data = self._dto_to_api(fields, _data)

                    # Mandar a bagatela por apis
                    self._api_client.enviar_dados(dict_data, entidade, self.api_key)

                    self._processar_blobs(entidade, service, dict_data)

                    # Aponta a leitura para a próxima página
                    _last = _data[-1]
                    current_after = getattr(_last, _last.pk_field)


                # Carrega os objetos pais a partir dos dados modificados nas entidades filhas
                lista_ids_entidades_atualizar = []
                for _chave, _campo_lista in service._dto_class.list_fields_map.items():
                    #_sub_dto      = _campo_lista.dto_type
                    _sub_entidade = _campo_lista.entity_type.table_name
                    _sub_data_ultima_integracao = entidades_pendentes_bd[_sub_entidade]
                    _campo_id_pai = _campo_lista.related_entity_field

                    #_coluna_id_filho = _sub_dto.fields_map[_sub_dto.pk_field].entity_field
                    para_apagar    = _dao.listar_dados_exclusao(_campo_id_pai, _sub_entidade, _sub_data_ultima_integracao)
                    para_atualizar = _dao.listar_dados_alteracao(_campo_id_pai, _sub_entidade, _sub_data_ultima_integracao)

                    lista_ids_entidades_atualizar = lista_ids_entidades_atualizar + para_apagar + para_atualizar


                if lista_ids_entidades_atualizar:

                    current_after = None
                    filtro_atualizacao = filters.copy() if filters else {}
                    filtro_atualizacao[_coluna_id] = ",".join([str(id) for id in lista_ids_entidades_atualizar if id is not None])

                    while True:

                        if self._interromper_execucao:
                            self._log("Processo interrompido pelo usuário.")
                            return

                        try:
                            _data = service.list(
                                    current_after,
                                    TAMANHO_PAGINA,
                                    fields,
                                    None,
                                    filtro_atualizacao,
                                    search_query=search_query,
                                )
                        except ProgrammingError as ex:
                            if self._is_missing_table_error(ex):
                                self._log(f"WARNING: {entidade}: tabela não encontrada no banco de dados. Pulando entidade.")
                                break
                            raise

                        _count = _count + len(_data)

                        if len(_data)==0:
                            break

                        self._log(f"{_count} registro(s).")
                        _resumo[entidade] = _resumo.get(entidade, 0) + _count

                        # Convertendo para o formato de dicionário (permitindo omitir campos do DTO) e add tenant
                        dict_data = self._dto_to_api(fields, _data)

                        # Mandar a bagatela por apis
                        self._api_client.enviar_dados(dict_data, entidade, self.api_key)

                        # Aponta a leitura para a próxima página
                        _last = _data[-1]
                        current_after = getattr(_last, _last.pk_field)


                self._atualiza_ultima_integracao(entidade, _filtros_integracao_cfg)

            self._log("")
            self._log(self._color("Integração finalizada com sucesso!", 92, _is_console))
            if _resumo:
                self._log("")
                self._log(self._color(f"Resumo da integração: {', '.join(f'{k}: {v}' for k, v in _resumo.items())}", 92, _is_console))

        finally:
            self._em_execucao = False


    @medir_tempo("Integração")
    def executar_integracao(self, parar_caso_erros: bool = False):
        self._interromper_execucao = False
        self._em_execucao = True

        inicio = time.perf_counter()
        _idx = 0
        _exclusoes: int = 0
        _atualizacoes: int = 0
        correlation_id = str(uuid.uuid4())

        _erros_envio = []

        try:
            assert self._integracao_foi_configurada(), "Integração não configurada!"

            _dao = self._integracao_dao()

            assert _dao.existem_grupos_empresariais_integracao_ativos(), "Nenhum grupo empresarial ativo para integração."
            assert _dao.listar_dados_particionamento(), "Entidades empresariais inválidas para integração."

            self._log(f"Executando integração para o Tenant: {self.tenant} .")

            # @TODO usar uma forma de se evitar o long pooling sem usar triggers...
            entidades_pendentes_bd = _dao.listar_entidades_pendentes_integracao()

            # Não filtrar entidades filhas
            entidades_pendentes = {
                entidade: entidades_pendentes_bd[entidade]
                    for entidade in self._entidades_integracao() if entidade in entidades_pendentes_bd.keys()
            }

            self._log(f"{len(entidades_pendentes)} entidades para processar." if entidades_pendentes else "Nenhuma entidade para processar.")
            _resumo = {}
            _resumo_exclusoes = {}
            _resumo_atualizacoes = {}

            if not entidades_pendentes:
                return

            self._integracao_dir = f"integracao_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}"

            _data_integracao: Dict[str, datetime.datetime] = {}

            _total_passos = len(entidades_pendentes) * 2

            self._telemetria_inicio_integracao(correlation_id, len(entidades_pendentes))

            # Processa exclusões gerais
            self._log("")
            self._log("Iniciando integração das exclusões pendentes.")
            for entidade, data_ultima_integracao in reversed(entidades_pendentes.items()):

                # if not entidade in ['ns.enderecos']:
                #     continue

                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return

                _idx += 1
                if not self._menos_log:
                    self._log(f"{entidade}, {_idx} de {_total_passos}.")

                service = self._injector.service_for(entidade, True)

                _data_integracao[entidade] = datetime.datetime.now(self._tz_br)

                _coluna_id = service._dto_class.fields_map[service._dto_class.pk_field].entity_field
                # TODO: ajustar para o horário do servidor. Fiz esse recursos devido a diferenã de timestamp entre o rastro e o registro,
                #  o que fazia com que o rastros ficasse atrasados em relação aos registros e consequentemente não houvesse exclusão.
                data_ultima_integracao_ajustada = data_ultima_integracao - datetime.timedelta(hours=3)
                #print(data_ultima_integracao_ajustada)
                para_apagar = _dao.listar_dados_exclusao(_coluna_id, entidade, data_ultima_integracao_ajustada)
                if para_apagar:

                    for i in range(0, len(para_apagar), TAMANHO_PAGINA):
                        bloco_ids = para_apagar[i:i+TAMANHO_PAGINA]
                        _resumo[entidade] = _resumo.get(entidade, 0) + len(bloco_ids)
                        _resumo_exclusoes[entidade] = _resumo_exclusoes.get(entidade, 0) + len(bloco_ids)
                        self._apagar_dados_bulk(bloco_ids, entidade, parar_caso_erros)
                        _exclusoes += len(bloco_ids)

                    if not self._menos_log:
                        self._log("Entidade integrada com sucesso.")
                    self._log(f"{entidade}: {len(para_apagar)} registro(s).")
                else:
                    if not self._menos_log:
                        self._log("Sem dados para atualizar, indo adiante.")



            # Processa upserts
            self._log("")
            self._log("Iniciando integração das atualizações pendentes.")
            for entidade, data_ultima_integracao in entidades_pendentes.items():

                # if not entidade in ['ns.enderecos']:
                #     continue

                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return

                _idx += 1
                if not self._menos_log:
                    self._log(f"{entidade}, {_idx} de {_total_passos}.")
                _count = 0

                # Carregar dados paginados para integrar
                service = self._injector.service_for(entidade, True)
                last_current_after = None
                current_after = None
                fields = self._fields_to_load_erp(service)
                _filtros_particionamento = self._filtro_particionamento_de(entidade)
                _filtros_integracao_cfg = self._filtros_integracao_cfg(entidade)
                filters = _filtros_particionamento | _cfg_filtros_to_dto_filtro(_filtros_integracao_cfg)
                search_query = None


                # Dados alterados apos data_ultima_integracao
                filtro_atualizacao = filters.copy() if filters else {}
                filtro_atualizacao['lastupdate'] = data_ultima_integracao
                while True:

                    if self._interromper_execucao:
                        self._log("Processo interrompido pelo usuário.")
                        return

                    _data = []
                    try:
                        _data = service.list(
                                current_after,
                                TAMANHO_PAGINA,
                                fields,
                                None,
                                filtro_atualizacao,
                                search_query=search_query,
                            )

                    except AfterRecordNotFoundException as e:
                        current_after = last_current_after
                        continue

                    except ProgrammingError as ex:
                        if self._is_missing_table_error(ex):
                            self._log(f"WARNING: {entidade}: tabela não encontrada no banco de dados. Pulando entidade.")
                            break
                        raise

                    _count = _count + len(_data)

                    if len(_data)==0:
                        break

                    # if not self._menos_log:
                    #     self._log(f"{entidade}: {_count} registro(s)")

                    _resumo[entidade] = _resumo.get(entidade, 0) + _count
                    _resumo_atualizacoes[entidade] = _resumo_atualizacoes.get(entidade, 0) + _count
                    _atualizacoes += _count

                    # Convertendo para o formato de dicionário (permitindo omitir campos do DTO) e add tenant
                    dict_data = self._dto_to_api(fields, _data)

                    # Mandar a bagatela por apis
                    _erros_envio += self._enviar_dados(dict_data, entidade, parar_caso_erros)

                    self._processar_blobs(entidade, service, dict_data, True, parar_caso_erros)

                    # Aponta a leitura para a próxima página
                    _last = _data[-1]
                    last_current_after = current_after
                    current_after = getattr(_last, _last.pk_field)


                # Carrega os objetos pais a partir dos dados modificados nas entidades filhas
                lista_ids_entidades_atualizar = []
                for _chave, _campo_lista in service._dto_class.list_fields_map.items():
                    _sub_entidade = _campo_lista.entity_type.table_name
                    _sub_data_ultima_integracao = entidades_pendentes_bd.get(_sub_entidade, data_ultima_integracao)
                    _campo_id_pai = _campo_lista.related_entity_field

                    para_apagar    = _dao.listar_dados_exclusao(_campo_id_pai, _sub_entidade, _sub_data_ultima_integracao)
                    para_atualizar = _dao.listar_dados_alteracao(_campo_id_pai, _sub_entidade, _sub_data_ultima_integracao)

                    lista_ids_entidades_atualizar = list(set(lista_ids_entidades_atualizar + para_apagar + para_atualizar))


                if lista_ids_entidades_atualizar:
                    self._log("Integrando modificações de entidades agregadas.")

                    last_current_after = None
                    current_after = None
                    filtro_atualizacao = filters.copy() if filters else {}
                    _coluna_id = service._dto_class.fields_map[service._dto_class.pk_field].entity_field

                    # A cada TAMANHO_PAGINA carrega um bloco de ids para evitar problemas de tamanho de query
                    for i in range(0, len(lista_ids_entidades_atualizar), TAMANHO_PAGINA):

                        bloco_ids = lista_ids_entidades_atualizar[i:i+TAMANHO_PAGINA]
                        # carrega os objetos filhos a partir dos ids do pai, isto é sua composição
                        filtro_atualizacao[_coluna_id] = ",".join([str(id) for id in bloco_ids if id is not None])

                        while True:

                            if self._interromper_execucao:
                                self._log("Processo interrompido pelo usuário.")
                                return

                            _data = []
                            try:

                                _data = service.list(
                                        current_after,
                                        TAMANHO_PAGINA,
                                        fields,
                                        None,
                                        filtro_atualizacao,
                                        search_query=search_query,
                                    )

                            except AfterRecordNotFoundException as e:
                                current_after = last_current_after
                                continue

                            except ProgrammingError as ex:
                                if self._is_missing_table_error(ex):
                                    self._log(f"WARNING: {entidade}: tabela não encontrada no banco de dados. Pulando entidade.")
                                    break
                                raise

                            _count = _count + len(_data)

                            if len(_data)==0:
                                break

                            # if not self._menos_log:
                            #     self._log(f"{entidade}: {_count} registro(s)")
                            _resumo[entidade] = _resumo.get(entidade, 0) + _count
                            _resumo_atualizacoes[entidade] = _resumo_atualizacoes.get(entidade, 0) + _count
                            _atualizacoes += _count

                            # Convertendo para o formato de dicionário (permitindo omitir campos do DTO) e add tenant
                            dict_data = self._dto_to_api(fields, _data)

                            # Mandar a bagatela por apis
                            _erros_envio += self._enviar_dados(dict_data, entidade, parar_caso_erros)

                            self._processar_blobs(entidade, service, dict_data, True, parar_caso_erros)

                            # Aponta a leitura para a próxima página
                            _last = _data[-1]
                            last_current_after = current_after
                            current_after = getattr(_last, _last.pk_field)

                if  _count == 0:
                    if not self._menos_log:
                        self._log("Sem dados para atualizar, indo adiante.")
                else:
                    if not self._menos_log:
                        self._log("Entidade integrada com sucesso.")
                    self._log(f"{entidade}: {_count} registro(s).")

                self._atualiza_data_ultima_integracao(entidade, _data_integracao[entidade], _filtros_integracao_cfg)

                self._telemetria_integracao_entidade(
                    correlation_id,
                    entidade,
                    _idx,
                    _resumo_exclusoes.get(entidade,0),
                    _resumo_atualizacoes.get(entidade,0),
                    _data_integracao[entidade]
                )

            self._log("")
            self._log(self._color("Integração finalizada com sucesso!", 92, _is_console))

            if _erros_envio:
                self._log(self._color("Ocorreram erros que foram ignorados durante o processo, verifique:", 91, _is_console))
                self._log("")
                self._log("\n"+("-"*80)+"\n"+self._color("\n".join(str(e)+ "\n" + "-"*80 for e in _erros_envio), 91, _is_console))

            if _resumo:
                self._log("")
                self._log(self._color(f"Resumo da integração: {', '.join(f'{k}: {v}' for k, v in _resumo.items())}", 92, _is_console))

        finally:
            fim = time.perf_counter()
            self._telemetria_fim_integracao(
                correlation_id,
                _idx,
                _atualizacoes,
                _exclusoes,
                (fim - inicio) * 1000,
                self._status_execucao(len(_erros_envio), parar_caso_erros, self._interromper_execucao)
            )
            self._em_execucao = False


    def integrity_fields(self, dto) -> dict:
        fields = {"root": set()}

        for _field_name in sorted(dto.integrity_check_fields_map.keys()):

            if _field_name in self._ignored_fields:
                continue

            _field_obj = dto.integrity_check_fields_map[_field_name]

            if isinstance(_field_obj, DTOField):
                fields["root"].add(_field_name)
                continue

            if isinstance(_field_obj, DTOListField):
                fields["root"].add(_field_name)
                fields.setdefault(_field_name, set())

                for _related_field in sorted(_field_obj.dto_type.integrity_check_fields_map.keys()):
                    if not _related_field in self._ignored_fields:
                        fields["root"].add(f"{_field_name}.{_related_field}")
                        fields[_field_name].add(_related_field)

        return fields


    def tratar_campos_comparacao(self, dados: dict, campos_ignorados: list):

        keys_to_delete = []
        for chave, valor in dados.items():

            # Remove timezone para comparação
            if isinstance(valor, (datetime.datetime, datetime.date)):
                if valor.tzinfo is not None:
                    dados[chave] = valor.astimezone(self._tz_br).replace(microsecond=0, tzinfo=None)
                else:
                    dados[chave] = valor.replace(microsecond=0, tzinfo=None)


            if "created_by" in dados and not dados["created_by"] is None:
                if not isinstance(dados["created_by"], dict):
                    try:
                        _value_dict = json_loads(dados["created_by"])
                        dados["created_by"] = _value_dict
                    except (TypeError, ValueError, JsonLoadException):
                        dados["created_by"] = {"id": dados["created_by"]}

            if "updated_by" in dados and not dados["updated_by"] is None:
                if not isinstance(dados["updated_by"], dict):
                    try:
                        _value_dict = json_loads(dados["updated_by"])
                        dados["updated_by"] = _value_dict
                    except (TypeError, ValueError, JsonLoadException):
                        dados["updated_by"] = {"id": dados["updated_by"]}

            # Ignora campos não úteis
            if chave in campos_ignorados:
                keys_to_delete.append(chave)

            # Aplica regras em sublistas
            if isinstance(valor, list):
                valor.sort(key=lambda x: x['id'])
                for item in valor:
                    self.tratar_campos_comparacao(item, campos_ignorados)

        for chave in keys_to_delete:
            del dados[chave]


    def converte_dados_para_hash(self, dto, integrity_fields):

        data = dto.convert_to_dict(integrity_fields)

        self.tratar_campos_comparacao(data, self._ignored_fields)

        concatenated_values = json_dumps(data)

        data['tenant'] = self.tenant

        return {
            'id': str(data[dto.pk_field]),
            'hash': hashlib.sha256(concatenated_values.encode('utf-8')).hexdigest(),
            '_source': data,
            '_source_hash': concatenated_values
        }

    def comparar_ids(self, dados_referencia, dados_comparacao):

        if dados_referencia['registros'] != dados_comparacao['registros']:
            self._log(self._color(f"Existem diferenças nas quantidades de dados:\r\n\r\nLocal: {dados_referencia['registros']}\r\n\r\nWeb  : {dados_comparacao['registros']}", 91, _is_console))


        # Índices para facilitar busca por ID (mantendo compatibilidade com a saída esperada do método)
        idx_referencia = {item['id']: item for item in dados_referencia['dados']}
        idx_comparacao = {item: {'id':item} for item in dados_comparacao['dados']}


        # Inicializar listas de mudanças
        _criar = []
        _excluir = []

        # Verificar itens nos dados de referência
        for item_id, item_ref in idx_referencia.items():
            if item_id not in idx_comparacao:
                # Criar se não existe nos dados de comparação
                _criar.append(item_ref['_source'])

        # Verificar itens nos dados de comparação
        for item_id in idx_comparacao.keys():
            if item_id not in idx_referencia:
                # Excluir se não existe em A
                _excluir.append(idx_comparacao[item_id]['id'])

        return _criar, _excluir


    def comparar_dados(self, dados_referencia, dados_comparacao):

        if dados_referencia['campos']['_'] != dados_comparacao['campos']['_']:
            self._log(self._color(f"Existem diferenças entre os campos comparados:\r\n\r\nLocal: {dados_referencia['campos']['_']}\r\n\r\nWeb  : {dados_comparacao['campos']['_']}", 91, _is_console ))

        if dados_referencia['registros'] != dados_comparacao['registros']:
            self._log(self._color(f"Existem diferenças nas quantidades de dados:\r\n\r\nLocal: {dados_referencia['registros']}\r\n\r\nWeb  : {dados_comparacao['registros']}", 91, _is_console))

        # Índices para facilitar busca por ID
        idx_referencia = {item['id']: item for item in dados_referencia['dados']}
        idx_comparacao = {item['id']: item for item in dados_comparacao['dados']}

        # Inicializar listas de mudanças
        _criar = []
        _atualizar = []
        _excluir = []
        _diff:List[tuple] = []

        # Verificar itens nos dados de referência
        for item_id, item_ref in idx_referencia.items():
            if item_id not in idx_comparacao:
                # Criar se não existe nos dados de comparação
                _criar.append(item_ref['_source'])
            elif item_ref['hash'] != idx_comparacao[item_id]['hash']:
                # Atualizar se o hash é diferente
                _atualizar.append(item_ref['_source'])
                # Adiciona para exibir os dados puros se disponível
                if '_source' in idx_comparacao[item_id]:
                    a = json_loads(item_ref['_source_hash'])
                    b = json_loads(idx_comparacao[item_id]['_source'])
                    _diff.append((a,b))

        # Verificar itens nos dados de comparação
        for item_id in idx_comparacao.keys():
            if item_id not in idx_referencia:
                # Excluir se não existe em A
                _excluir.append(idx_comparacao[item_id]['id'])

        return _criar, _atualizar, _excluir, _diff


    def _log_integridade(self, msg):
        _du.save_to_file(f'{self._integridade_dir}/log_diferencas_integridade.log', msg)


    @property
    def last_integridade_log_path(self) -> str:
        """Retorna o caminho absoluto do arquivo de log da última verificação de integridade executada.
        Retorna None se nenhuma verificação foi executada ainda."""
        if not hasattr(self, '_integridade_dir') or not self._integridade_dir:
            return None
        return os.path.join(os.getcwd(), self._integridade_dir, 'log_diferencas_integridade.log')


    def _color(self, text, code, console):
        if console:
            return f"\033[{code}m{text}\033[0m"
        else:
            return text


    def _log_comparacao_objetos(self, id, obj1, obj2, caminho='', console=False):
        _out = self._log if console else self._log_integridade

        if isinstance(obj1, dict) and isinstance(obj2, dict):
            for k in set(obj1.keys()).union(obj2.keys()):
                self._log_comparacao_objetos(id, obj1.get(k), obj2.get(k), f"{caminho}.{k}" if caminho else k)
        elif isinstance(obj1, list) and isinstance(obj2, list):
            max_len = max(len(obj1), len(obj2))
            for i in range(max_len):
                item1 = obj1[i] if i < len(obj1) else None
                item2 = obj2[i] if i < len(obj2) else None
                _id = obj1[i]['id'] if not item1 is None else obj2[i]['id'] if not item2 is None else 'indefinido'
                self._log_comparacao_objetos(id, item1, item2, f"{caminho}[{_id}]")
        else:
            s1 = str(obj1)
            s2 = str(obj2)
            if s1 != s2:
                s1_pad = s1.ljust(25)
                s2_pad = s2.ljust(25)
                _id = str(id)

                _out(f"{_id:<40} {caminho:<40} {self._color(s1_pad, '31', console)} {self._color(s2_pad, '32', console)}")


    def _log_diferencas(self, entidade, data_ultima_integracao, console=False):
        _out = self._log if console else self._log_integridade
        _out("\r\n")
        _out("-" * 130)
        _out(f"Entidade: {self._color(entidade, '36', console)}")
        _out(f"Data da última integração: {self._color(data_ultima_integracao, '36', console)}")
        _out(f"{'ID':<40} {'Campo':<40} {'Local':<25} {'Nuvem':<25}")
        _out("-" * 130)


    def _log_cabecalho(self, entidade, data_ultima_integracao, console=False):
        _out = self._log if console else self._log_integridade
        _out("\r\n")
        _out("=" * 130)
        _out(f"Entidade: {self._color(entidade, '36', console)}")
        _out(f"Data da última integração: {self._color(data_ultima_integracao, '36', console)}")


    def _log_grupo(self, titulo, detalhado=False, console=False):
        _out = self._log if console else self._log_integridade
        _out("=" * 130)
        _out(f"{titulo}:")
        _out(f"{'ID':<40} {'Campo':<40} {'Local':<25} {'Nuvem':<25}" if detalhado else "ID")
        _out("=" * 130)


    def _log_id(self, _id, console=False):
        _out = self._log if console else self._log_integridade
        _out(f"{_id}")


    @medir_tempo("Verificação de integridade")
    def executar_verificacao_integridade(
        self,
        entidades: list,
        tipo_verificacao: TipoVerificacaoIntegridade = TipoVerificacaoIntegridade.HASH,
        parar_caso_diferencas : bool = False,
        detalhar_diferencas: bool = False,
        corrigir_auto: bool = False,
        tenant: int = 0,
        trace: bool = False,
        verificacao_rapida: bool = False
    ):
        self._interromper_execucao = False
        self._em_execucao = True

        try:

            assert self._integracao_foi_configurada(), "Integração não configurada!"

            _dao = self._integracao_dao()

            assert _dao.existem_grupos_empresariais_integracao_ativos(), "Nenhum grupo empresarial ativo para integração."
            assert _dao.listar_dados_particionamento(), "Entidades empresariais inválidas para integração."

            self._log(f"Executando verificação de integridade para o Tenant: {self.tenant} .")

            self._detalhar_diferencas = detalhar_diferencas

            self._trace = trace

            if corrigir_auto:
                assert self.tenant==tenant, "Tenant informado para correção não é igual ao configurado"

            self._integridade_dir = f"verificacao_integridade_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}"

            # Remover entidades que nao devem ser processadas
            entidades_verificacao = copy.copy(self._entidades_integracao())
            if entidades:
                for entidade in entidades:
                    assert entidade in self._entidades_integracao(), f"Entidade '{entidade}' não consta como entidade para integração!"

                for entidade in self._entidades_integracao():
                    if not entidade in entidades:
                        entidades_verificacao.remove(entidade)


            self._log(f"{len(entidades_verificacao)} entidade(s) para verificar integridade.")

            _diferencas = False
            _idx = 0
            _resumo = defaultdict(list)
            _exclusoes_retentar = []
            _atualizacoes_retentar = []
            for entidade in entidades_verificacao:

                if  entidade in _ignorar_integridade:
                    self._log(f"Ignorando verificação de integridade para a entidade {entidade}.")
                    continue

                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return

                _idx += 1
                self._log(f"Verificando integridade {entidade}, {_idx} de {len(entidades_verificacao)}.")

                # Carregar dados paginados para integrar
                service = self._injector.service_for(entidade, True)

                _count = 0
                current_after = None
                fields = self._fields_to_load_erp(service)
                _filtros_particionamento = self._filtro_particionamento_de(entidade)
                _filtros_integracao_cfg = self._filtros_integracao_cfg(entidade)
                filters = (
                    _filtros_particionamento
                    | _cfg_filtros_to_dto_filtro(_filtros_integracao_cfg)
                )
                
                # Adiciona filtros de data se necessário
                if verificacao_rapida:
                    
                    data_ate = datetime.datetime.now(self._tz_br) + datetime.timedelta(days=1)
                    data_de = data_ate - datetime.timedelta(days=7)
                    
                    filters = filters | {"data_de": self._format_data(data_de), "data_ate": self._format_data(data_ate)}

                search_query = None
                _integrity_fields = self.integrity_fields(service._dto_class)
                _dados_locais = []

                self._log("Extraindo dados para comparação.")
                pular_entidade = False
                while True:

                    if self._interromper_execucao:
                        self._log("Processo interrompido pelo usuário.")
                        return

                    try:
                        if (tipo_verificacao is TipoVerificacaoIntegridade.IDENTIFICADOR) and not corrigir_auto:

                            _dao = self._injector.integridade_dao(service._entity_class)
                            _entity_filters = service._create_entity_filters(filters)
                            _data = _dao.list_ids_integridade(
                                after=current_after,
                                limit=TAMANHO_PAGINA,
                                order_fields=None,
                                filters=_entity_filters
                            )
                        else:

                            _data = service.list(
                                current_after,
                                TAMANHO_PAGINA,
                                fields,
                                None,
                                filters,
                                search_query=search_query,
                            )
                    except ProgrammingError as ex:
                        if self._is_missing_table_error(ex):
                            self._log(f"WARNING: {entidade}: tabela não encontrada no banco de dados. Pulando entidade.")
                            pular_entidade = True
                            break
                        raise

                    _count = _count + len(_data)

                    if len(_data)==0:
                        break

                    self._log(f"{_count} registro(s).")

                    # Aponta a leitura para a próxima página
                    _last = _data[-1]
                    if tipo_verificacao is TipoVerificacaoIntegridade.IDENTIFICADOR and not corrigir_auto:
                        current_after = _last
                    else:
                        current_after = getattr(_last, _last.pk_field)

                    # Convertendo para o formato de dicionário (permitindo omitir campos do DTO) e add tenant
                    if tipo_verificacao is TipoVerificacaoIntegridade.IDENTIFICADOR:

                        if corrigir_auto:
                            _data = self._dto_to_api(fields, _data)
                            while _data:
                                _item = _data.pop(0)
                                _dados_locais.append({
                                    'id': str(_item['id']),
                                    '_source': _item
                                })
                        else:
                            while _data:
                                _item = _data.pop(0)
                                _dados_locais.append({
                                    'id': str(_item),
                                    '_source': {'id' : _item }
                                })
                    else:

                        _cp_fields = copy.deepcopy(_integrity_fields)
                        while _data:
                            dto = _data.pop(0)
                            _dados_locais.append(self.converte_dados_para_hash(dto, _cp_fields))


                if pular_entidade:
                    continue

                if tipo_verificacao is TipoVerificacaoIntegridade.IDENTIFICADOR:
                    _campos = service._dto_class.pk_field
                else:
                    _campos = ",".join(sorted(_integrity_fields['root']))

                    #break

                _dados_locais = {
                    'registros' : _count,
                    'campos': {
                        "_": _campos,
                    },
                    'dados': _dados_locais
                }

                # captura os dados de integridade da entidade
                if self._web_data_service:
                    self._log("Consultando dados do banco Web (acesso direto).")
                else:
                    self._log("Consultando dados da API HTTP.")
                _dados = []
                _count = 0

                # Divide os filtros em lotes para evitar URLs muito longas (erro 414).
                # Cada lote tem seu próprio cursor de paginação (ultimo_id) independente,
                # evitando o erro 500 que ocorria ao usar um cursor de um lote em outro.
                if self._web_data_service:
                    lotes_filtros = [filters]  # WebDataService não sofre limite de URL
                else:
                    lotes_filtros = self._api_client._dividir_filtros_em_lotes(filters, tamanho_lote=20)

                for lote_filtros in lotes_filtros:
                    _ultimo_id = None  # cursor reiniciado a cada lote

                    while True:

                        if self._interromper_execucao:
                            self._log("Processo interrompido pelo usuário.")
                            return

                        try:
                            # Usa WebDataService se disponível, senão usa ApiClient
                            if self._web_data_service:
                                _dados_remotos = self._web_data_service.consultar_dados_integridade(
                                    entidade=entidade,
                                    campos=_campos,
                                    tipo_verificacao='ids' if tipo_verificacao is TipoVerificacaoIntegridade.IDENTIFICADOR else 'dados',
                                    filtros_cfg=_filtros_integracao_cfg,
                                    ultimo_id=_ultimo_id,
                                    limit=TAMANHO_PAGINA
                                )
                            else:
                                _dados_remotos = self._api_client.consultar_integridade_de(
                                    entidade, lote_filtros, _filtros_integracao_cfg,
                                    _ultimo_id, detalhar_diferencas, self.api_key, self.tenant,
                                    tipo_verificacao
                                )
                        except Exception as e:
                            self._log("")
                            self._log(f"\n{'-'*80}\nErro:\n{str(e)}\n{'-'*80}")
                            self._log("")
                            raise e

                        _count = _count + len(_dados_remotos['dados'])

                        if len(_dados_remotos['dados']) == 0:
                            break

                        self._log(f"{_count} registro(s).")

                        _dados = _dados + copy.copy(_dados_remotos['dados'])

                        # Cursor usa os dados da página atual (não o acumulado global),
                        # garantindo que seja um ID válido dentro do filtro do lote corrente.
                        if tipo_verificacao is TipoVerificacaoIntegridade.IDENTIFICADOR:
                            _ultimo_id = _dados_remotos['dados'][-1]
                        else:
                            _ultimo_id = _dados_remotos['dados'][-1]['id']

                _dados_remotos['dados'] = _dados
                _dados_remotos['registros'] = _count

                self._log("Comparando dados.")

                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return

                # Compara os dados e obtem o que se deve fazer
                para_criar, para_atualizar, para_apagar, _diff = [], [], [] ,[]

                if tipo_verificacao is TipoVerificacaoIntegridade.IDENTIFICADOR:
                    para_criar, para_apagar = self.comparar_ids(_dados_locais, _dados_remotos)
                else:
                    para_criar, para_atualizar, para_apagar, _diff = self.comparar_dados(_dados_locais, _dados_remotos)

                if para_criar or para_atualizar or para_apagar:
                    _resumo[entidade].append(self._color(f"Local: {_dados_locais['registros']}  Web: {_dados_remotos['registros']}", 93, _is_console))

                    _dt_ultima_integracao = self._dao_intg.data_ultima_integracao(entidade)
                    self._log_cabecalho(entidade, _dt_ultima_integracao, _is_console)

                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return

                if para_apagar:
                    _resumo[entidade].append(self._color(f"Para apagar -> {len(para_apagar)}", 93, _is_console))
                    self._log(self._color(f"\r\n{_resumo[entidade][-1]}\r\n", 93, _is_console))

                    self._log_grupo(f"Dados que serão excluídos: {len(para_apagar)}", False, _is_console)

                    if _diff:
                        for _desktop in para_apagar:
                            self._log_comparacao_objetos(_desktop, _desktop, None)
                    else:
                        for _desktop in para_apagar:
                            self._log_id(_desktop, _is_console)

                    if corrigir_auto:
                        self._log(f"\r\nRemovendo dados em {entidade}.\r\n")

                        for i in range(0, len(para_apagar), TAMANHO_PAGINA):
                            _dados_pagina = para_apagar[i:i+TAMANHO_PAGINA]

                            self._log(f"\r\nApagando página {i+1} de {len(para_apagar)-1}.\r\n")

                            try:
                                # Usa WebDataService se disponível, senão usa ApiClient
                                if self._web_data_service:
                                    ids_apagar = [item if isinstance(item, str) else item['id'] for item in _dados_pagina]
                                    self._web_data_service.deletar_dados_bulk(
                                        entidade=entidade,
                                        ids=ids_apagar
                                    )
                                else:
                                    self._api_client.apagar_dados_bulk(_dados_pagina, entidade, self.api_key, self.tenant)
                            except Exception:
                                _exclusoes_retentar.append((entidade, _dados_pagina))


                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return


                if para_criar:
                    _resumo[entidade].append(self._color(f"Para criar -> {len(para_criar)}", 93, _is_console))
                    self._log(self._color(f"\r\n{_resumo[entidade][-1]}\r\n", 93, _is_console))

                    self._log_grupo(f"Dados que serão criados: {len(para_criar)}", False, _is_console)

                    if _diff:
                        for _desktop in para_criar:
                            self._log_comparacao_objetos(_desktop['id'], _desktop, {})
                    else:
                        for _desktop in para_criar:
                            self._log_id(_desktop['id'], _is_console)

                    if corrigir_auto:
                        self._log(f"\r\nCriando dados em {entidade}.\r\n")

                        for i in range(0, len(para_criar), TAMANHO_PAGINA):
                            _dados_pagina = para_criar[i:i+TAMANHO_PAGINA]
                            try:
                                # Usa WebDataService se disponível, senão usa ApiClient
                                if self._web_data_service:
                                    self._web_data_service.upsert_dados(
                                        entidade=entidade,
                                        dados=_dados_pagina
                                    )
                                else:
                                    self._api_client.enviar_dados(_dados_pagina, entidade, self.api_key)
                            except Exception:
                                _atualizacoes_retentar.append((entidade, _dados_pagina))

                            self._processar_blobs(entidade, service, _dados_pagina, False)


                if self._interromper_execucao:
                    self._log("Processo interrompido pelo usuário.")
                    return


                if para_atualizar:
                    _resumo[entidade].append(self._color(f"Para atualizar -> {len(para_atualizar)}", 93, _is_console))
                    self._log(self._color(f"\r\n{_resumo[entidade][-1]}\r\n", 93, _is_console))

                    self._log_grupo(f"Dados que serão atualizados: {len(para_atualizar)}", _diff, _is_console)

                    if _diff:
                        _i : int = 0
                        for _desktop, _web in _diff:
                            _i  += 1
                            self._log_comparacao_objetos(_desktop['id'], _desktop, _web)
                            self._trace_check(f"{self._integridade_dir}/integridade_{entidade.replace('.','_')}_{_desktop['id']}_{_i}_LOCAL.txt", json_dumps(_desktop))
                            self._trace_check(f"{self._integridade_dir}/integridade_{entidade.replace('.','_')}_{_web['id']}_{_i}_REMOTE.txt", json_dumps(_web))
                    else:
                        for _desktop in para_atualizar:
                            self._log_id(_desktop['id'], _is_console)

                    if corrigir_auto:
                        self._log(f"\r\nAtualizando dados em {entidade}.\r\n")

                        for i in range(0, len(para_atualizar), TAMANHO_PAGINA):
                            _dados_pagina = para_atualizar[i:i+TAMANHO_PAGINA]
                            try:
                                # Usa WebDataService se disponível, senão usa ApiClient
                                if self._web_data_service:
                                    self._web_data_service.upsert_dados(
                                        entidade=entidade,
                                        dados=_dados_pagina
                                    )
                                else:
                                    self._api_client.enviar_dados(_dados_pagina, entidade, self.api_key)
                            except Exception:
                                _atualizacoes_retentar.append((entidade, _dados_pagina))

                            self._processar_blobs(entidade, service, _dados_pagina)


                if not _diferencas:
                    _diferencas = para_criar or para_atualizar or para_apagar

                if parar_caso_diferencas and (para_criar or para_atualizar or para_apagar) and not corrigir_auto:
                    break

            # Retentando correções que falharam em ordem reversa
            _erros_envio = []
            if _atualizacoes_retentar or _exclusoes_retentar:

                self._log("Enviando ajustes de dados que falharam na primeira tentativa.")

                for _entidade_retentativa, _dados_pagina in reversed(_exclusoes_retentar):
                    self._log(f"\r\nRemovendo dados em {_entidade_retentativa}.\r\n")
                    self._apagar_dados_bulk(_dados_pagina, _entidade_retentativa)

                for _entidade_retentativa, _dados_pagina in reversed(_atualizacoes_retentar):
                    self._log(f"\r\nCriando/Atualizando dados em {_entidade_retentativa}.\r\n")
                    _service = None
                    _erros_envio += self._enviar_dados(_dados_pagina, _entidade_retentativa)
                    self._processar_blobs(_entidade_retentativa, _service, _dados_pagina)


            if _diferencas:
                self._log(self._color("\r\nOcorreram diferenças na checagem da integridade, verifique a saída.\r\n", 93, _is_console))

            if not _diferencas:
                self._log(self._color("Verificação finalizada sem diferenças!\r\n", 92, _is_console))

            if _resumo:
                self._log(self._color("Resumo da integração:\r\n", 92, _is_console))
                for entidade, detalhes in _resumo.items():
                    self._log(f"{entidade}:  " + '\n'.join(detalhes) + "\n")

                if corrigir_auto:
                    self._log(self._color("Foram enviados dados de correção durante o processo, verifique a saída.\r\n",92, _is_console))

            if _erros_envio:
                self._log(self._color("Ocorreram erros que foram ignorados durante o processo, verifique:", 91, _is_console))
                self._log("")
                self._log("\n"+("-"*80)+"\n"+self._color("\n".join(str(e)+ "\n" + "-"*80 for e in _erros_envio), 91, _is_console))

        finally:
            self._em_execucao = False
            
    def _format_data(self, dt):                        
        if isinstance(dt, (datetime.datetime, datetime.date)):
            return dt.strftime("%Y-%m-%d %H:%M:%S.%f") if isinstance(dt, datetime.datetime) else dt.strftime("%Y-%m-%d 00:00:00.000000")
        return dt



"""
Métricas de performance para telemetria.

Este módulo implementa a captura automática de métricas de performance
como tempo de execução, uso de memória e CPU durante operações.
"""

import time
import psutil
import logging
from dataclasses import dataclass, field
from typing import Dict, Any, Optional, List
from contextlib import contextmanager


@dataclass
class MetricasPerformance:
    """Dados de métricas de performance."""

    operacao: str
    duracao_ms: float
    memoria_inicial_mb: float
    memoria_final_mb: float
    memoria_maxima_mb: float
    cpu_percentual: float
    timestamp_inicio: float
    timestamp_fim: float
    status: str = "sucesso"
    erro: Optional[str] = None
    dados_extras: Dict[str, Any] = field(default_factory=dict)


class TelemetriaMetrics:
    """
    Sistema de métricas de performance para telemetria.

    Captura automaticamente métricas de tempo, memória e CPU
    durante a execução de operações.
    """

    def __init__(self):
        """Inicializa o sistema de métricas."""
        self.logger = logging.getLogger(__name__)
        self._metricas: List[MetricasPerformance] = []
        self._processo = psutil.Process()

    def _obter_uso_memoria_mb(self) -> float:
        """
        Obtém o uso de memória em MB.

        Returns:
            Uso de memória em MB
        """
        try:
            return self._processo.memory_info().rss / 1024 / 1024
        except Exception as e:
            self.logger.warning("Erro ao obter uso de memória: %s", str(e))
            return 0.0

    def _obter_uso_cpu_percentual(self) -> float:
        """
        Obtém o uso de CPU em percentual.

        Returns:
            Uso de CPU em percentual
        """
        try:
            return self._processo.cpu_percent(interval=0.1)
        except Exception as e:
            self.logger.warning("Erro ao obter uso de CPU: %s", str(e))
            return 0.0

    @contextmanager
    def monitorar_operacao(self, nome_operacao: str,
                          dados_extras: Optional[Dict[str, Any]] = None):
        """
        Context manager para monitorar uma operação.

        Args:
            nome_operacao: Nome da operação a ser monitorada
            dados_extras: Dados extras a serem incluídos nas métricas

        Yields:
            Objeto com métricas da operação
        """
        timestamp_inicio = time.time()
        memoria_inicial = self._obter_uso_memoria_mb()
        cpu_inicial = self._obter_uso_cpu_percentual()

        metricas = MetricasPerformance(
            operacao=nome_operacao,
            duracao_ms=0.0,
            memoria_inicial_mb=memoria_inicial,
            memoria_final_mb=memoria_inicial,
            memoria_maxima_mb=memoria_inicial,
            cpu_percentual=cpu_inicial,
            timestamp_inicio=timestamp_inicio,
            timestamp_fim=timestamp_inicio,
            dados_extras=dados_extras or {}
        )

        try:
            yield metricas

            # Operação concluída com sucesso
            metricas.status = "sucesso"

        except Exception as e:
            # Operação falhou
            metricas.status = "erro"
            metricas.erro = str(e)
            self.logger.error("Erro na operação %s: %s", nome_operacao, str(e))
            raise

        finally:
            # Finalizar métricas
            timestamp_fim = time.time()
            memoria_final = self._obter_uso_memoria_mb()

            metricas.timestamp_fim = timestamp_fim
            metricas.duracao_ms = (timestamp_fim - timestamp_inicio) * 1000
            metricas.memoria_final_mb = memoria_final
            metricas.memoria_maxima_mb = max(memoria_inicial, memoria_final)

            # Adicionar às métricas coletadas
            self._metricas.append(metricas)

            self.logger.debug(
                "Operação %s concluída: %.2fms, memória: %.2fMB",
                nome_operacao, metricas.duracao_ms, metricas.memoria_maxima_mb
            )

    def obter_metricas(self) -> List[MetricasPerformance]:
        """
        Obtém todas as métricas coletadas.

        Returns:
            Lista de métricas coletadas
        """
        return self._metricas.copy()


# Instância global para uso em decoradores
_instancia_metricas = TelemetriaMetrics()


def obter_metricas() -> List[MetricasPerformance]:
    """
    Obtém todas as métricas coletadas.

    Returns:
        Lista de métricas coletadas
    """
    return _instancia_metricas.obter_metricas()


import datetime
import uuid

def serialize_value(value):
    if isinstance(value, (uuid.UUID, datetime.datetime)):
        return str(value)
    elif isinstance(value, list):
        return [serialize_value(v) for v in value]
    elif isinstance(value, dict):
        return {k: serialize_value(v) for k, v in value.items()}
    return value

def set_nested(d, keys, value):
    """Define o valor em dicionário aninhado baseado nos `keys`"""
    current = d
    for key in keys[:-1]:
        if isinstance(current, list):
            # Neste caso, aplicamos a lógica em cada item da lista
            for item in current:
                set_nested(item, keys[keys.index(key):], value)
            return
        current = current.setdefault(key, {})
    if isinstance(current, list):
        # Aplica valor em todos os itens da lista
        for item in current:
            item[keys[-1]] = serialize_value(value)
    else:
        current[keys[-1]] = serialize_value(value)


def construir_objeto(campos, valores):
    resultado = {}
    campos_base_completos = set()

    # Primeiro, identificar campos que já têm uma lista inteira (ex: itensfaixas)
    for campo, valor in zip(campos, valores):
        if '.' not in campo and isinstance(valor, list):
            resultado[campo] = serialize_value(value=valor)
            campos_base_completos.add(campo)

    # Agora processar os campos normais e subcampos
    for campo, valor in zip(campos, valores):
        if '.' in campo:
            partes = campo.split('.')
            base = partes[0]
            if base in campos_base_completos:
                continue  # ignora subcampos de listas já preenchidas
            if base not in resultado:
                resultado[base] = []
            set_nested(resultado, partes, valor)
        elif campo not in resultado:
            resultado[campo] = serialize_value(valor)

    return resultado


def color(text, code):
    return f"\033[{code}m{text}\033[0m"

def comparar(v1, v2, caminho=''):
    if isinstance(v1, dict) and isinstance(v2, dict):
        for k in set(v1.keys()).union(v2.keys()):
            comparar(v1.get(k), v2.get(k), f"{caminho}.{k}" if caminho else k)
    elif isinstance(v1, list) and isinstance(v2, list):
        max_len = max(len(v1), len(v2))
        for i in range(max_len):
            item1 = v1[i] if i < len(v1) else None
            item2 = v2[i] if i < len(v2) else None
            comparar(item1, item2, f"{caminho}[{v1[i]['id'][:8]}]")
    else:
        s1 = str(v1)
        s2 = str(v2)
        if s1 != s2:
            s1_pad = s1.ljust(25)
            s2_pad = s2.ljust(25)
            print(f"{caminho:<40} {color(s1_pad, '31')} {color(s2_pad, '32')} {color('≠', '33')}")

def mostrar_diferencas(entidade, id, linha1, linha2):
    print(f"ID: {color(id,'36')}")
    print(f"Entidade: {color(entidade,'36')}")
    print(f"{'Campo':<40} {'Local':<25} {'Nuvem':<25} {'Status'}")
    print("-" * 100)

    comparar(linha1, linha2)


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'resources/form_v2/config_inicial.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ConfigInicial(object):
    def setupUi(self, ConfigInicial):
        ConfigInicial.setObjectName("ConfigInicial")
        ConfigInicial.resize(600, 450)
        self.verticalLayout = QtWidgets.QVBoxLayout(ConfigInicial)
        self.verticalLayout.setObjectName("verticalLayout")
        self.labelStatus = QtWidgets.QLabel(ConfigInicial)
        self.labelStatus.setObjectName("labelStatus")
        self.verticalLayout.addWidget(self.labelStatus)
        self.labelChave = QtWidgets.QLabel(ConfigInicial)
        self.labelChave.setObjectName("labelChave")
        self.verticalLayout.addWidget(self.labelChave)
        self.lineEditChave = QtWidgets.QLineEdit(ConfigInicial)
        self.lineEditChave.setObjectName("lineEditChave")
        self.verticalLayout.addWidget(self.lineEditChave)
        self.labelGrupos = QtWidgets.QLabel(ConfigInicial)
        self.labelGrupos.setObjectName("labelGrupos")
        self.verticalLayout.addWidget(self.labelGrupos)
        self.scrollArea = QtWidgets.QScrollArea(ConfigInicial)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.checkBoxGrupoA = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.checkBoxGrupoA.setObjectName("checkBoxGrupoA")
        self.verticalLayout_2.addWidget(self.checkBoxGrupoA)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea)
        self.btnValidar = QtWidgets.QPushButton(ConfigInicial)
        self.btnValidar.setObjectName("btnValidar")
        self.verticalLayout.addWidget(self.btnValidar)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btnCancelar = QtWidgets.QPushButton(ConfigInicial)
        self.btnCancelar.setObjectName("btnCancelar")
        self.horizontalLayout.addWidget(self.btnCancelar)
        self.btnSalvar = QtWidgets.QPushButton(ConfigInicial)
        self.btnSalvar.setObjectName("btnSalvar")
        self.horizontalLayout.addWidget(self.btnSalvar)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(ConfigInicial)
        QtCore.QMetaObject.connectSlotsByName(ConfigInicial)

    def retranslateUi(self, ConfigInicial):
        _translate = QtCore.QCoreApplication.translate
        ConfigInicial.setWindowTitle(_translate("ConfigInicial", "Configuração da Integração por APIs"))
        self.labelStatus.setText(_translate("ConfigInicial", "STATUS: Integração não configurada"))
        self.labelChave.setText(_translate("ConfigInicial", "Chave de Ativação:"))
        self.lineEditChave.setPlaceholderText(_translate("ConfigInicial", "AAAA-BBBB-CCCC-DDDD"))
        self.labelGrupos.setText(_translate("ConfigInicial", "Grupos Empresariais Disponíveis:"))
        self.checkBoxGrupoA.setText(_translate("ConfigInicial", "Grupo A - Matriz"))
        self.btnValidar.setText(_translate("ConfigInicial", "Validar Chave"))
        self.btnCancelar.setText(_translate("ConfigInicial", "Cancelar"))
        self.btnSalvar.setText(_translate("ConfigInicial", "Salvar Configuração"))

def show_form():
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ConfigInicial = QtWidgets.QDialog()
    ui = Ui_ConfigInicial()
    ui.setupUi(ConfigInicial)
    ConfigInicial.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    show_form()




from typing import List
from datetime import datetime
from nsj_gcf_utils.db_adapter2 import DBAdapter2
import uuid

class JobDAO:

    _db: DBAdapter2

    def __init__(self, db: DBAdapter2):
        self._db = db

    def get_job_type_by_code(self, codigo):
        """
        Retorna o JobType correspondente ao CODIGO passado.
        """

        sql = """
            select
                codigo as codigo,
                funcao as funcao,
                modoexecucao as modo_execucao,
                schema as schema,
                sistema as sistema,
                urlscript as url_script,
                pacote as pacote,
                classe as classe,
                versao as versao
            from
                util.jobtypes
            where
                codigo = :codigo
        """

        return self._db.execute_query_first_result(sql, codigo=codigo)


    def cria_job_type(self, sistema, modoexecucao, codigo):

        sql = """
            insert into util.jobtypes (
            codigo,
            sistema,
            modoexecucao
            ) values (
            :codigo,
            :sistema,
            :modoexecucao
            ) returning
                codigo as codigo,
                funcao as funcao,
                modoexecucao as modo_execucao,
                schema as schema,
                sistema as sistema,
                urlscript as url_script,
                pacote as pacote,
                classe as classe,
                versao as versao
        """

        rowcount, _ = self._db.execute(sql, codigo=codigo, sistema=sistema, modoexecucao=modoexecucao)

        if rowcount!=1:
            raise Exception(
                "Erro ao registrar grupos empresariais no banco de dados"
            )


    def get_agendamento_integracao(self):

        sql = """
            select
                a.jobschedule as id,
                b.codigo,
                a.descricao,
                a.jobtype,
                a.entrada,
                a.tipoagendamento,
                a.intervaloagendamento as intervalo,
                a.status
            from util.jobschedule a
            join util.jobtypes b on ( b.jobtype = a.jobtype )
            where
                b.codigo = 'INTEGRACAO_APIS' and
                a.tenant = 0 and
                a.tipoagendamento = 1
        """

        return self._db.execute_query_first_result(sql)


    def agenda_job(self, entrada, intervalo):

        # tipoagendamento -
        # 0 - Agendado para todos os dias, no horário do campo "agendamento" (ignorando o dia do mesmo campo);
        # 1 - Agendado para executar a cada X minutos (de acordo com o campo intervaloagendamento);
        # 2 - Agendado para o momento definido no campo agendamento;

        sql = """
            INSERT INTO util.jobschedule (
                codigo,
                descricao,
                jobtype,
                entrada,
                tipoagendamento,
                intervaloagendamento
            ) VALUES (
                'INTEGRACAO_NASAJON',
                'Piloto do Job de integração por apis para substituir o symmetric',
                (SELECT jobtype FROM util.jobtypes WHERE codigo='INTEGRACAO_APIS'),
                :entrada,
                1,
                :intevalo --intervalo de execucao (minutos)
            );
        """
        rowcount, _ = self._db.execute(sql, entrada=entrada, intevalo=intervalo)

        if rowcount!=1:
            raise Exception(
                "Erro ao registrar agendamento no banco de dados"
            )


    def atualiza_job(self, agendamento, entrada, intervalo):

        # tipoagendamento -
        # 0 - Agendado para todos os dias, no horário do campo "agendamento" (ignorando o dia do mesmo campo);
        # 1 - Agendado para executar a cada X minutos (de acordo com o campo intervaloagendamento);
        # 2 - Agendado para o momento definido no campo agendamento;

        sql = """
            UPDATE util.jobschedule
            SET
                status = 0,
                intervaloagendamento = :intevalo,
                entrada = :entrada
            WHERE
                jobschedule = :agendamento;
        """
        rowcount, _ = self._db.execute(sql, agendamento=agendamento, entrada=entrada, intevalo=intervalo)

        if rowcount!=1:
            raise Exception(
                "Erro ao registrar agendamento no banco de dados"
            )


    def cancela_agendamento(self, agendamento):
        sql = """
            UPDATE util.jobschedule
            SET STATUS = 3
            WHERE jobschedule = :jobschedule
        """
        rowcount, _ = self._db.execute(sql, jobschedule=agendamento)

        if rowcount!=1:
            raise Exception(
                "Erro ao excluir agendamento no banco de dados"
            )


import sys
from PyQt5.QtWidgets import (
    QApplication, QMainWindow, QVBoxLayout, QHBoxLayout, QWidget,
    QLabel, QComboBox, QPushButton, QTableWidget, QTableWidgetItem, QCheckBox
)

class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Gestão de Eventos - PyQt5")
        self.setGeometry(100, 100, 800, 400)

        # Layout principal
        main_layout = QVBoxLayout()

        # Topo: seleção de ambiente e botões
        top_layout = QHBoxLayout()
        top_layout.addWidget(QLabel("Ambiente:"))
        self.ambiente_combo = QComboBox()
        self.ambiente_combo.addItems(["Homologação", "Produção"])
        self.ambiente_combo.setCurrentText("Homologação")
        top_layout.addWidget(self.ambiente_combo)

        # Botões no topo
        for btn_text in ["Configurar", "Enviar Eventos", "Consultar Lotes"]:
            top_layout.addWidget(QPushButton(btn_text))

        main_layout.addLayout(top_layout)

        # Tabela
        self.table = QTableWidget(0, 6)  # 0 linhas, 6 colunas
        self.table.setHorizontalHeaderLabels(
            ["Empresa", "Grupo", "Evento", "Identificação", "Data/Hora Geração", "Enviar?"]
        )
        main_layout.addWidget(self.table)

        # Botão Enviar
        enviar_button = QPushButton("Enviar Eventos Selecionados")
        enviar_button.clicked.connect(self.enviar_eventos)
        main_layout.addWidget(enviar_button)

        # Preenchendo a tabela
        dados = [
            ("003", "Inicial/Tabela", "S-1010", "Cadastro Rubrica A", "06/08/2024 15:32:02"),
            ("003", "Não Periódico", "S-2200", "Admissão Funcionário 222", "20/09/2024 09:46:34"),
        ]
        self.preencher_tabela(dados)

        # Configurar o layout principal
        central_widget = QWidget()
        central_widget.setLayout(main_layout)
        self.setCentralWidget(central_widget)

    def preencher_tabela(self, dados):
        self.table.setRowCount(len(dados))
        for row_idx, (empresa, grupo, evento, identificacao, data_hora) in enumerate(dados):
            self.table.setItem(row_idx, 0, QTableWidgetItem(empresa))
            self.table.setItem(row_idx, 1, QTableWidgetItem(grupo))
            self.table.setItem(row_idx, 2, QTableWidgetItem(evento))
            self.table.setItem(row_idx, 3, QTableWidgetItem(identificacao))
            self.table.setItem(row_idx, 4, QTableWidgetItem(data_hora))

            # Checkbox na coluna "Enviar?"
            checkbox = QCheckBox()
            checkbox.setStyleSheet("margin-left:50%; margin-right:50%;")  # Centralizar o checkbox
            self.table.setCellWidget(row_idx, 5, checkbox)

    def enviar_eventos(self):
        selecionados = []
        for row_idx in range(self.table.rowCount()):
            checkbox = self.table.cellWidget(row_idx, 5)
            if checkbox and checkbox.isChecked():
                evento = self.table.item(row_idx, 2).text()  # Obter o valor da coluna "Evento"
                selecionados.append(evento)

        print("Eventos selecionados para envio:", selecionados)

def show_form():
    import sys
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    show_form()


import time
import json
from typing import Iterator
from importlib.metadata import version

import requests
from requests.models import Response

from nsj_gcf_utils.json_util import convert_to_dumps, json_dumps

from nsj_integracao_api_client.service.integrador_cfg import (
    Environment, TAMANHO_PAGINA, _E_SEND_DATA, AUTH_HEADER,
    TipoVerificacaoIntegridade
)

from nsj_integracao_api_client.infra.debug_utils import DebugUtils as _du

class APIException(Exception):
    _campos_padrao = ["mensagem", "endpoint",  "status", "resposta"]
    _todos_campos =  ["mensagem", "endpoint", "status", "resposta", "dados"]

    def __init__(self, mensagem=None, endpoint=None, status=None, reason=None, response_text=None, dados=None):
        super().__init__(mensagem)
        self.mensagem = mensagem
        self.endpoint = endpoint
        self.status = status
        self.reason = reason
        self.response_text = response_text
        self.dados = dados
        self._campos_visiveis = self._campos_padrao.copy()

    def campos(self, **kwargs):
        # kwargs: campo=True/False
        self._campos_visiveis = [campo for campo, mostrar in kwargs.items() if mostrar and campo in self._todos_campos]
        return self

    def __str__(self):
        campos_dict = {
            "mensagem": self.mensagem,
            "endpoint": self.endpoint,
            "status": f"{self.status} - {self.reason}",
            "resposta": self.response_text,
            "dados": self.dados
        }
        return "\n".join(f"{campo}: {campos_dict.get(campo)}" for campo in self._campos_visiveis if campos_dict.get(campo) is not None)


class ApiClient:


    def __init__(self, env: Environment):
        self._env = env


    def _trace_envio(self, filename, content):
        _du.conditional_trace(
            condition=_E_SEND_DATA,
            func=_du.save_to_file,
            filename=filename,
            content=content
        )


    def _url_base(self) -> str:
        if self._env == Environment.LOCAL:
            return "http://localhost:5000/integracao-pessoas-api/66"
        elif self._env == Environment.DEV:
            return "https://api.nasajon.dev/integracao-pessoas-api/66"
        elif self._env == Environment.QA:
            return "https://api.nasajon.qa/integracao-pessoas-api/66"
        elif self._env == Environment.PROD:
            return "https://api4.nasajon.app/integracao-pessoas-api/66"
        else:
            raise ValueError(f"Ambiente desconhecido: {self._env}")


    def _url_diretorio(self) -> str:
        if self._env == Environment.LOCAL:
            return "http://localhost"
        elif self._env == Environment.DEV:
            return "https://dir.nasajon.dev"
        elif self._env == Environment.QA:
            return "https://dir.nasajon.qa"
        elif self._env == Environment.PROD:
            return "https://diretorio.nasajon.com.br"
        else:
            raise ValueError(f"Ambiente desconhecido: {self._env}")


    def raise_exception_from_response(
        self,
        msg: str,
        response: Response,
        dados: list,
        response_text: str = None
    ):
        response_content = response_text if response_text else response.json() if 'application/json' in response.headers.get('Content-Type', '') else response.text

        raise APIException(
            msg,
            response.url,
            response.status_code,
            response.reason,
            response_content,
            dados
        )


    def sistemas_contratados_ambiente(self, api_key: str, tenant: int):
        """
        Endpoint que efetua a consulta dos sistemas contratados para um ambiente (tenant).
        """

        assert api_key, "Autenticação não foi informada."
        assert tenant, "Ambiente não foi informado."

        s = requests.Session()
        s.headers.update({
            'apiKey': api_key,
            'Accept':'application/json'
        })
        response = s.get(f'{self._url_diretorio()}/v2/api/licenca/tenant/{tenant}')

        if response.status_code == 200:
            try:
                return response.json()
            except Exception as ex:
                raise APIException(f'Retorno desconhecido:{response.text}') from ex

        if response.status_code < 200 or response.status_code > 299:
            if 'application/json' in response.headers.get('Content-Type', ''):
                _json_response = response.json()
                _message = _json_response['erro'] if 'erro' in _json_response else ''
            else:
                _message = response.text
            raise APIException(
                "Erro ao capturar os sistemas contratados",
                response.url,
                response.status_code,
                response.reason,
                _message
            )


    def enviar_dados(self, dict_data: list, entidade: str, api_key: str):
        """
        """
        self._trace_envio(
            f"trace/send_data_{entidade}_{_du.time()}.json",
            json.dumps(dict_data, indent=2, ensure_ascii=False, default=str) if _E_SEND_DATA  else ""
        )

        _rota = entidade.replace('.', '/').lower()

        upsert = True #False if (acao in ["processos"] and self._env == Environment.PROD) else True

        s = requests.Session()
        s.headers.update({'Content-Type':'application/json', AUTH_HEADER: api_key})

        if upsert:
            response = s.put(f'{self._url_base()}/{_rota}?upsert=true', json=convert_to_dumps(dict_data))

            if response.status_code == 413:
                for _item in dict_data:
                    response = s.put(f'{self._url_base()}/{_rota}?upsert=true', json=convert_to_dumps(_item))
                    self._tratar_resposta(response, dict_data)
            else:
                self._tratar_resposta(response, dict_data)

        else:
            for _item in dict_data:
                response = s.post(f'{self._url_base()}/{_rota}', json=convert_to_dumps(_item))
                if response.status_code < 200 or response.status_code > 299:
                    if 'application/json' in response.headers.get('Content-Type', '') and \
                    isinstance(response.json(), list) and 'message' in response.json()[0] and \
                    ('_bt_check_unique' in response.json()[0].get('message', '') or response.status_code == 409):
                        raise APIException(
                            "Erro ao enviar dados ao servidor",
                            response.url,
                            response.status_code,
                            response.reason,
                            response.text,
                            convert_to_dumps(dict_data)
                        )


    def apagar_dados(self, dict_data: list, entidade: str, api_key: str, tenant: int) -> Iterator[str]:
        """
        """
        _rota = entidade.replace('.', '/').lower()

        s = requests.Session()
        s.headers.update({'Content-Type':'application/json', AUTH_HEADER: api_key})
        response = s.delete(f'{self._url_base()}/{_rota}?tenant={tenant}', json=convert_to_dumps(dict_data))

        # Caso algum item não exista no servidor tenta apagar individualmente,
        # ignorando os ausentes
        if response.status_code == 404:
            #Não mais faremos essa tratativa do um a um
            for _item in dict_data:
                response = s.delete(f'{self._url_base()}/{_rota}?tenant={tenant}', json=convert_to_dumps([_item]))
                if response.status_code == 404:
                    #yield f"Registro {entidade} id: {_item} não encontrado para exclusão no tenant {tenant}, ignorando."
                    continue
                else:
                    if (response.status_code < 200 or response.status_code > 299) and response.status_code != 404:
                        if 'application/json' in response.headers.get('Content-Type', ''):
                            _json_response = response.json()
                            if isinstance(_json_response, dict):
                                _message = _json_response['message']if 'message' in _json_response else ''
                            else:
                                _message = _json_response[0]['message'] if 'message' in _json_response[0] else ''

                        else:
                            _message = response.text
                        raise APIException(
                            "Erro ao apagar dados do servidor",
                            response.url,
                            response.status_code,
                            response.reason,
                            _message,
                            convert_to_dumps(dict_data)
                        )

        if (response.status_code < 200 or response.status_code > 299) and response.status_code != 404:
            if 'application/json' in response.headers.get('Content-Type', ''):
                _json_response = response.json()
                if isinstance(_json_response, dict):
                    _message = _json_response['message']if 'message' in _json_response else ''
                else:
                    _message = _json_response[0]['message'] if 'message' in _json_response[0] else ''

            else:
                _message = response.text
            raise APIException(
                "Erro ao apagar dados do servidor",
                response.url,
                response.status_code,
                response.reason,
                _message,
                convert_to_dumps(dict_data)
            )

        return []


    def _validar_response_apagar_dados_bulk(self, response: Response) -> bool:

        # Valida o tipo de response
        if not 'application/json' in response.headers.get('Content-Type', ''):
            return False

        # Valida o formato de response
        _json_response = response.json()
        if  not (isinstance(_json_response, dict) and 'global_status' in _json_response  and 'response' in _json_response):
            return False

        return True


    def apagar_dados_bulk(self, data: list, entidade: str, api_key: str, tenant: int):
        """
        """
        _rota = entidade.replace('.', '/').lower()

        s = requests.Session()
        s.headers.update({'Content-Type':'application/json', AUTH_HEADER: api_key})
        response = s.delete(f'{self._url_base()}/{_rota}/bulk?tenant={tenant}', json=convert_to_dumps(data))

        if not self._validar_response_apagar_dados_bulk(response):
            self.raise_exception_from_response("Formato inválido na resposta da exclusão em massa", response, data)

        _json_response = response.json()

        # Avalia os estatus individuais para lançar erro se pertinente
        if _json_response["global_status"] != "OK":

            _responses = [ f'{_resp["error"]["message"]}' for _resp in _json_response["response"] if _resp["status"] >= 500 ]

            if _responses:
                self.raise_exception_from_response(
                    "Erro ao apagar dados do servidor",
                    response,
                    data,
                    "\n".join(_responses)
                )


    def _dividir_filtros_em_lotes(self, filtros: dict, tamanho_lote: int = 20):
        """
        Divide filtros que são listas/strings longas em lotes menores para evitar URI muito longa.
        Retorna uma lista de dicts de filtros, cada um representando um lote.
        """
        if not filtros:
            return [{}]

        # Identifica filtros que podem ser grandes (valores que são listas ou strings com vírgulas)
        filtro_grande = None
        filtro_grande_chave = None
        outros_filtros = {}

        for chave, valor in filtros.items():
            # Se for uma string com vírgulas (múltiplos valores concatenados)
            if isinstance(valor, str) and ',' in valor:
                filtro_grande = valor.split(',')
                filtro_grande_chave = chave
            else:
                outros_filtros[chave] = valor

        # Se não há filtro grande, retorna como está
        if not filtro_grande:
            return [filtros]

        # Divide o filtro grande em lotes
        lotes = []
        for i in range(0, len(filtro_grande), tamanho_lote):
            lote_valores = filtro_grande[i:i + tamanho_lote]
            lote_filtros = outros_filtros.copy()
            lote_filtros[filtro_grande_chave] = ','.join(lote_valores)
            lotes.append(lote_filtros)

        return lotes

    def consultar_integridade_de(
            self,
            entidade: str,
            filtros: dict,
            filtros_extras: list,
            ultimo_id,
            detalhar_diferencas: bool,
            api_key: str,
            tenant: int,
            tipo_verificacao: TipoVerificacaoIntegridade = TipoVerificacaoIntegridade.HASH):
        """
        Consulta uma página de dados de integridade na API, para os filtros e cursor (ultimo_id)
        informados. Não realiza paginação nem batching internamente — isso é responsabilidade
        do chamador (ver integrador.py).
        """
        _rota = entidade.replace('.', '/').lower()

        s = requests.Session()
        s.headers.update({
            'Content-Type': 'application/json',
            AUTH_HEADER: api_key,
        })

        filtros_str = None
        if filtros:
            filtros_str = "&".join(
                [f"{_chave}={filtros[_chave]}" for _chave in filtros.keys()]
            )

        filtros_extras_str = None
        if filtros_extras:
            filtros_extras_str = f"custom_filter={json_dumps(filtros_extras)}"
        else:
            filtros_extras_str = "custom_filter=[]"

        _url = (
            f'{self._url_base()}/{_rota}/verificacao-integridade?tenant={tenant}&source={detalhar_diferencas}'
            f'{"&" + filtros_str if filtros_str else ""}'
            f'&limit={TAMANHO_PAGINA}'
            f'{"&after="+str(ultimo_id) if ultimo_id else ""}'
            f'{"&" + filtros_extras_str}'
            f'&type={tipo_verificacao.value}'
        )

        _max_retries = 3
        _retry_count = 0
        while _retry_count <= _max_retries:
            response = s.get(_url, headers={'X-NSJ-INTEGRACAO-VERSION': version('nsj_integracao_api_entidades')})
            response_content = response.json() if 'application/json' in response.headers.get('Content-Type', '') else response.text

            if response.status_code < 200 or response.status_code > 299:
                _retry_count += 1
                if _retry_count > _max_retries:
                    if isinstance(response_content, dict):
                        _message = response_content.get('message', '')
                    else:
                        _message = response_content
                    raise APIException(
                        "Erro ao consultar a integridade no servidor",
                        response.url,
                        response.status_code,
                        response.reason,
                        _message
                    )
                else:
                    # Tenta novamente após um pequeno delay
                    time.sleep(2 ** _retry_count)
                    continue
            break

        return response_content


    def gerar_token_tenant(self, chave_ativacao: str) -> str:
        """
        """
        s = requests.Session()
        s.headers.update({
            'Content-Type':'application/x-www-form-urlencoded',
            'Accept':'application/json'
        })
        response = s.post(
            f'{self._url_diretorio()}/v2/api/gerar_token_ativacao_integracao/',
            data={"codigo_ativacao": chave_ativacao})

        if response.status_code == 200:
            _json = response.json()
            if "apiKey" in _json:
                return _json["apiKey"]
            else:
                raise APIException(f'Retorno desconhecido:{_json}')

        if response.status_code < 200 or response.status_code > 299:
            if 'application/json' in response.headers.get('Content-Type', ''):
                _json_response = response.json()
                _message = _json_response['message'] if 'message' in _json_response else ''
            else:
                _message = response.text
            raise APIException(
                "Erro ao gerar o token de integração",
                response.url,
                response.status_code,
                response.reason,
                _message
            )

    def ativar_integracao(self, tenant: int, chave_ativacao: str, grupos_empresariais: list, api_key: str) -> dict:
        """
        Ativa integração para um tenant usando a API.
        
        Args:
            tenant: ID do tenant
            chave_ativacao: Chave de ativação no formato NNNN-NNNN-NNNN-NNNN
            grupos_empresariais: Lista de códigos ou UUIDs dos grupos empresariais
            api_key: API Key para autenticação
            
        Returns:
            dict: Resposta da API com detalhes da ativação
            
        Raises:
            APIException: Em caso de erro na chamada da API
        """
        s = requests.Session()
        s.headers.update(
            {
                "Content-Type": "application/json",
                "Accept": "application/json",
                AUTH_HEADER: api_key,
            }
        )
        
        payload = {
            "tenant": tenant,
            "chave_ativacao": chave_ativacao,
            "grupos_empresariais": grupos_empresariais,
        }
        
        response = s.post(
            f"{self._url_base()}/processos-admin/ativacao-integracao",
            json=payload,
        )
        
        if response.status_code == 201:
            return response.json()
        
        # Tratamento de erros
        if "application/json" in response.headers.get("Content-Type", ""):
            _json_response = response.json()
            _message = _json_response.get("erro", _json_response.get("message", ""))
            _detalhes = _json_response.get("mensagem", _json_response.get("detalhes", ""))
            if _detalhes:
                _message = f"{_message}: {_detalhes}"
        else:
            _message = response.text
            
        raise APIException(
            "Erro ao ativar integração",
            response.url,
            response.status_code,
            response.reason,
            _message,
        )

    def ativar_grupos_empresariais(self, tenant: int, grupos: list, api_key: str) -> dict:
        """
        Ativa grupos empresariais para integração via API.
        
        Args:
            tenant: ID do tenant
            grupos: Lista de códigos ou UUIDs dos grupos empresariais
            api_key: API Key para autenticação
            
        Returns:
            dict: Resposta da API com detalhes da ativação
            
        Raises:
            APIException: Em caso de erro na chamada da API
        """
        s = requests.Session()
        s.headers.update(
            {
                "Content-Type": "application/json",
                "Accept": "application/json",
                AUTH_HEADER: api_key,
            }
        )
        
        payload = {
            "tenant": tenant,
            "grupos": grupos,
        }
        
        response = s.post(
            f"{self._url_base()}/processos-admin/grupos-empresariais",
            json=payload,
        )
        
        if response.status_code == 200:
            return response.json()
        
        # Tratamento de erros
        if "application/json" in response.headers.get("Content-Type", ""):
            _json_response = response.json()
            _message = _json_response.get("erro", _json_response.get("message", ""))
        else:
            _message = response.text
            
        raise APIException(
            "Erro ao ativar grupos empresariais",
            response.url,
            response.status_code,
            response.reason,
            _message,
        )

    def desativar_grupos_empresariais(self, tenant: int, grupos: list, api_key: str) -> dict:
        """
        Desativa grupos empresariais da integração via API.
        
        Args:
            tenant: ID do tenant
            grupos: Lista de códigos ou UUIDs dos grupos empresariais
            api_key: API Key para autenticação
            
        Returns:
            dict: Resposta da API com detalhes da desativação
            
        Raises:
            APIException: Em caso de erro na chamada da API
        """
        s = requests.Session()
        s.headers.update(
            {
                "Content-Type": "application/json",
                "Accept": "application/json",
                AUTH_HEADER: api_key,
            }
        )
        
        payload = {
            "tenant": tenant,
            "grupos": grupos,
        }
        
        response = s.post(
            f"{self._url_base()}/processos-admin/grupos-empresariais/desativar",
            json=payload,
        )
        
        if response.status_code == 200:
            return response.json()
        
        # Tratamento de erros
        if "application/json" in response.headers.get("Content-Type", ""):
            _json_response = response.json()
            _message = _json_response.get("erro", _json_response.get("message", ""))
        else:
            _message = response.text
            
        raise APIException(
            "Erro ao desativar grupos empresariais",
            response.url,
            response.status_code,
            response.reason,
            _message,
        )

    def desativar_integracao(self, tenant: int, api_key: str) -> dict:
        """
        Desativa integração para um tenant via API (remove token).
        
        Args:
            tenant: ID do tenant
            api_key: API Key para autenticação
            
        Returns:
            dict: Resposta da API
            
        Raises:
            APIException: Em caso de erro na chamada da API
        """
        s = requests.Session()
        s.headers.update(
            {
                "Accept": "application/json",
                AUTH_HEADER: api_key,
            }
        )
        
        response = s.delete(
            f"{self._url_base()}/processos-admin/ativacao-integracao?tenant={tenant}"
        )
        
        if response.status_code == 200:
            return response.json()
        
        # Tratamento de erros
        if "application/json" in response.headers.get("Content-Type", ""):
            _json_response = response.json()
            _message = _json_response.get("erro", _json_response.get("message", ""))
            _detalhes = _json_response.get("mensagem", "")
            if _detalhes:
                _message = f"{_message}: {_detalhes}"
        else:
            _message = response.text
            
        raise APIException(
            "Erro ao desativar integração",
            response.url,
            response.status_code,
            response.reason,
            _message,
        )

    def _tratar_resposta(self, response, dict_data):
        if response.status_code < 200 or response.status_code > 299:
            if 'application/json' in response.headers.get('Content-Type', ''):
                _json_response = response.json()
                if isinstance(_json_response, dict):
                    _message = _json_response['message'] if 'message' in _json_response else ''
                else:
                    _message = _json_response[0]['message'] if 'message' in _json_response[0] else ''
            else:
                _message = response.text
            raise APIException(
                "Erro ao enviar dados ao servidor",
                response.url,
                response.status_code,
                response.reason,
                _message,
                convert_to_dumps(dict_data)
            )


    def consultar_hash_blob(self, ids: list, entidade: str, campo: str, tenant: int, api_key: str):
        """
        """

        _rota = entidade.split('.', 1)[1].lower()

        s = requests.Session()
        s.headers.update({'Content-Type':'application/json', AUTH_HEADER: api_key})
        _url = (
            f'{self._url_base()}/blobs-{_rota}-hashes/filtros?tenant={tenant}&field={campo}'
            f'&limit={TAMANHO_PAGINA}'
        )

        _max_retries = 3
        _retry_count = 0
        while _retry_count <= _max_retries:
            response = s.post(_url, json=convert_to_dumps(ids))
            response_content = response.json() if 'application/json' in response.headers.get('Content-Type', '') else response.text

            if response.status_code < 200 or response.status_code > 299:

                _retry_count += 1
                if _retry_count > _max_retries:
                    if isinstance(response_content, dict):
                        _message = response_content.get('message', '')
                    else:
                        _message = response_content
                    raise APIException(
                        "Erro ao consultar os hashes dos blobs no servidor",
                        response.url,
                        response.status_code,
                        response.reason,
                        _message
                    )
                else:
                    # Tenta novamente após um pequeno delay
                    time.sleep(2 ** _retry_count)
                    continue
            break

        if not 'result' in response_content:
            raise APIException(
                "Formato inválido na consulta dos hashes dos blobs no servidor",
                response.url,
                response.status_code,
                response.reason,
                response_content
            )

        return response_content


    def enviar_blobs(self, ids: list, files: list ,entidade: str, campo: str, tenant: int, api_key: str):
        """
        """

        _rota = entidade.split('.', 1)[1].lower()

        s = requests.Session()
        s.headers.update({AUTH_HEADER: api_key})
        _url = f'{self._url_base()}/blobs-{_rota}?field={campo}&tenant={tenant}'

        _files = files  # só os arquivos
        _data = {'ids': json.dumps(ids)}

        response = s.put(_url, files=_files, data=_data)

        if response.status_code == 413:
            # Envia cada arquivo individualmente em caso de erro 413 (Payload Too Large)
            for i, file in enumerate(files):
                _data = {'ids': json.dumps([ids[i]])}
                response = s.put(_url, files=[file], data=_data)
                self._tratar_resposta(response, [ids[i]])
        else:
            self._tratar_resposta(response, ids)


# if __name__ == "__main__":
#     api = ApiClient(Environment.LOCAL)
#     key = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJ0aXBvIjoidGVuYW50IiwidGVuYW50X2lkIjo5Njd9.jZrhVlEjPYaffhsuZ9UiE0kMoXpVpfrvRTfNx0zdJGk"
#     files = [
#         ('files', ("None", b"file1", 'application/octet-stream'))
#     ]
#     api.enviar_blobs(["6a2dfb51-12a0-4716-aa90-97d333d44f09"], files, "persona.trabalhadores", "foto", 900, key)


"""
Decoradores para captura automática de eventos de telemetria.

Este módulo implementa decoradores que permitem capturar automaticamente
eventos de telemetria no início e fim de funções, ou eventos simples.
"""

import functools
import logging
import time
from typing import Callable, Optional
from nsj_integracao_api_client.infra.telemetria.telemetria_service import TelemetriaService
from nsj_integracao_api_client.infra.telemetria.telemetria_metrics import TelemetriaMetrics


def telemetria(evento_inicio: str, evento_fim: Optional[str] = None):
    """
    Decorador para capturar eventos de telemetria no início e fim de uma função.

    Args:
        evento_inicio: Código do evento de início
        evento_fim: Código do evento de fim (opcional, usa evento_inicio + "_FIM" se não fornecido)

    Returns:
        Decorador da função
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Obter instâncias
            telemetria_service = TelemetriaService()
            metrics = TelemetriaMetrics()

            # Determinar evento de fim
            evento_fim_final = evento_fim or f"{evento_inicio}_FIM"

            # Dados do evento de início
            dados_inicio = {
                "funcao": func.__name__,
                "modulo": func.__module__,
                "timestamp_inicio": time.time()
            }

            # Enviar evento de início
            try:
                telemetria_service._enviar_evento(
                    evento=evento_inicio,
                    resultado=f"Iniciando execução da função {func.__name__}",
                    dados_resultado=dados_inicio
                )
            except Exception as e:
                logging.getLogger(__name__).warning(
                    "Erro ao enviar evento de início %s: %s", evento_inicio, str(e)
                )

            # Executar função com monitoramento de métricas
            try:
                with metrics.monitorar_operacao(func.__name__) as metricas:
                    resultado = func(*args, **kwargs)
                    metricas.dados_extras["resultado_sucesso"] = True
                    return resultado

            except Exception as e:
                # Capturar erro nas métricas
                with metrics.monitorar_operacao(func.__name__) as metricas:
                    metricas.status = "erro"
                    metricas.erro = str(e)
                    metricas.dados_extras["resultado_sucesso"] = False
                    raise

            finally:
                # Dados do evento de fim
                dados_fim = {
                    "funcao": func.__name__,
                    "modulo": func.__module__,
                    "timestamp_fim": time.time(),
                    "duracao_ms": metricas.duracao_ms if 'metricas' in locals() else 0,
                    "memoria_utilizada_mb": metricas.memoria_maxima_mb if 'metricas' in locals() else 0,
                    "status": metricas.status if 'metricas' in locals() else "desconhecido"
                }

                # Enviar evento de fim
                try:
                    telemetria_service._enviar_evento(
                        evento=evento_fim_final,
                        resultado=f"Finalizando execução da função {func.__name__}",
                        dados_resultado=dados_fim
                    )
                except Exception as e:
                    logging.getLogger(__name__).warning(
                        "Erro ao enviar evento de fim %s: %s", evento_fim_final, str(e)
                    )

        return wrapper
    return decorator


def telemetria_simples(evento: str):
    """
    Decorador para capturar um evento simples de telemetria.

    Args:
        evento: Código do evento a ser enviado

    Returns:
        Decorador da função
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Obter instância do serviço
            telemetria_service = TelemetriaService()

            # Dados do evento
            dados_evento = {
                "funcao": func.__name__,
                "modulo": func.__module__,
                "timestamp": time.time(),
                "args_count": len(args),
                "kwargs_count": len(kwargs)
            }

            # Executar função
            try:
                resultado = func(*args, **kwargs)

                # Adicionar informações de sucesso
                dados_evento["status"] = "sucesso"
                dados_evento["resultado_tipo"] = type(resultado).__name__

                # Enviar evento
                try:
                    telemetria_service._enviar_evento(
                        evento=evento,
                        resultado=f"Execução bem-sucedida da função {func.__name__}",
                        dados_resultado=dados_evento
                    )
                except Exception as e:
                    logging.getLogger(__name__).warning(
                        "Erro ao enviar evento %s: %s", evento, str(e)
                    )

                return resultado

            except Exception as e:
                # Adicionar informações de erro
                dados_evento["status"] = "erro"
                dados_evento["erro"] = str(e)
                dados_evento["tipo_erro"] = type(e).__name__

                # Enviar evento de erro
                try:
                    telemetria_service._enviar_evento(
                        evento=evento,
                        resultado=f"Erro na execução da função {func.__name__}",
                        dados_resultado=dados_evento
                    )
                except Exception as telemetria_error:
                    logging.getLogger(__name__).warning(
                        "Erro ao enviar evento %s: %s", evento, str(telemetria_error)
                    )

                # Re-levantar o erro original
                raise

        return wrapper
    return decorator


# pylint: disable=W0212
import unittest
from unittest.mock import MagicMock
from nsj_integracao_api_client.service.integrador import IntegradorService
from nsj_integracao_api_client.service.integrador import IntegradorService
from nsj_integracao_api_client.infra.injector_factory import InjectorFactory
from nsj_integracao_api_client.dao.integracao import IntegracaoDAO

class TestLogComparacaoObjetos(unittest.TestCase):
    def setUp(self):
        self.log = MagicMock()

        # mock dao
        self._dao = MagicMock(spec=IntegracaoDAO)
        self._dao.recuperar_configuracao_ambiente.return_value = "PROD"
        self._dao.recuperar_token.return_value = ""

        # mock service
        self._injector = MagicMock(spec=InjectorFactory)
        self._injector.integracao_dao.return_value = self._dao

        self.integrador = IntegradorService(self._injector, self.log)
        self.integrador._log_integridade = MagicMock()
        self.integrador._color = lambda text, code, console: text

    def teste_comparacao_dicionarios(self):
        obj1 = {"a": 1, "b": 2}
        obj2 = {"a": 1, "b": 3}
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_comparacao_listas(self):
        obj1 = [{"id": 1, "value": "A"}, {"id": 2, "value": "B"}]
        obj2 = [{"id": 1, "value": "A"}, {"id": 2, "value": "C"}]
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_lista_vazia(self):
        obj1 = []
        obj2 = [{"id": 1, "value": "A"}]
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_lista_diferentes(self):
        obj1 = [{"id": 1, "value": "A"}]
        obj2 = [{"id": 1, "value": "A"},{"id": 2, "value": "A"}]
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_tipos_diferentes(self):
        obj1 = {"a": 1}
        obj2 = [{"id": 1, "value": "A"}]
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_estruturas_aninhadas(self):
        obj1 = {"a": [{"id": 1, "value": "A"}]}
        obj2 = {"a": [{"id": 1, "value": "B"}]}
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_estruturas_aninhadas_diferentes_a(self):
        obj1 = {"a": []}
        obj2 = {"a": [{"id": 1, "value": "B"}]}
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_estruturas_aninhadas_diferentes_b(self):
        obj1 = {"c":"c", "a": [{"id": 1, "value": "B"}]}
        obj2 = {"c":"c", "a": []}
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

    def teste_estruturas_aninhadas_diferentes_c(self):
        obj1 = {"c":"c1", "a": []}
        obj2 = {"c":"c2", "a": [{"id": 1}]}
        self.integrador._log_comparacao_objetos("test_id", obj1, obj2)
        self.integrador._log_integridade.assert_called()

if __name__ == "__main__":
    unittest.main()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'resources/form/pendencias_detalhes.ui'
#
# Created by: PyQt5 UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_FormItensPendentesDetalhes(object):
    def setupUi(self, FormItensPendentesDetalhes):
        FormItensPendentesDetalhes.setObjectName("FormItensPendentesDetalhes")
        FormItensPendentesDetalhes.resize(1024, 768)
        FormItensPendentesDetalhes.setMinimumSize(QtCore.QSize(1024, 768))
        FormItensPendentesDetalhes.setMaximumSize(QtCore.QSize(1024, 768))
        self.verticalLayout = QtWidgets.QVBoxLayout(FormItensPendentesDetalhes)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(FormItensPendentesDetalhes)
        self.label.setStyleSheet("font-size: 16px; font-weight: bold; margin: 4px;")
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.edtFiltro = QtWidgets.QLineEdit(FormItensPendentesDetalhes)
        self.edtFiltro.setClearButtonEnabled(True)
        self.edtFiltro.setObjectName("edtFiltro")
        self.verticalLayout.addWidget(self.edtFiltro)
        self.groupItensDetalhes = QtWidgets.QGroupBox(FormItensPendentesDetalhes)
        self.groupItensDetalhes.setTitle("")
        self.groupItensDetalhes.setObjectName("groupItensDetalhes")
        self.verticalLayoutGrupos = QtWidgets.QVBoxLayout(self.groupItensDetalhes)
        self.verticalLayoutGrupos.setObjectName("verticalLayoutGrupos")
        self.tableItensDetalhes = QtWidgets.QTableWidget(self.groupItensDetalhes)
        self.tableItensDetalhes.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectColumns)
        self.tableItensDetalhes.setCornerButtonEnabled(False)
        self.tableItensDetalhes.setObjectName("tableItensDetalhes")
        self.tableItensDetalhes.setColumnCount(2)
        self.tableItensDetalhes.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.tableItensDetalhes.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableItensDetalhes.setHorizontalHeaderItem(1, item)
        self.verticalLayoutGrupos.addWidget(self.tableItensDetalhes)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.btn_prev = QtWidgets.QPushButton(self.groupItensDetalhes)
        self.btn_prev.setObjectName("btn_prev")
        self.horizontalLayout.addWidget(self.btn_prev)
        self.lbl_page = QtWidgets.QLabel(self.groupItensDetalhes)
        self.lbl_page.setObjectName("lbl_page")
        self.horizontalLayout.addWidget(self.lbl_page)
        self.btn_next = QtWidgets.QPushButton(self.groupItensDetalhes)
        self.btn_next.setObjectName("btn_next")
        self.horizontalLayout.addWidget(self.btn_next)
        self.verticalLayoutGrupos.addLayout(self.horizontalLayout)
        self.verticalLayout.addWidget(self.groupItensDetalhes)

        self.retranslateUi(FormItensPendentesDetalhes)
        QtCore.QMetaObject.connectSlotsByName(FormItensPendentesDetalhes)

    def retranslateUi(self, FormItensPendentesDetalhes):
        _translate = QtCore.QCoreApplication.translate
        FormItensPendentesDetalhes.setWindowTitle(_translate("FormItensPendentesDetalhes", "Integração por APIs"))
        self.label.setText(_translate("FormItensPendentesDetalhes", "<Entidade>"))
        self.edtFiltro.setPlaceholderText(_translate("FormItensPendentesDetalhes", "Digite para filtrar...\""))
        item = self.tableItensDetalhes.horizontalHeaderItem(0)
        item.setText(_translate("FormItensPendentesDetalhes", "ID"))
        item = self.tableItensDetalhes.horizontalHeaderItem(1)
        item.setText(_translate("FormItensPendentesDetalhes", "Dados"))
        self.btn_prev.setText(_translate("FormItensPendentesDetalhes", "⬅ Anterior"))
        self.lbl_page.setText(_translate("FormItensPendentesDetalhes", "Página 0 de 0"))
        self.btn_next.setText(_translate("FormItensPendentesDetalhes", "Próximo ➡"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    FormItensPendentesDetalhes = QtWidgets.QWidget()
    ui = Ui_FormItensPendentesDetalhes()
    ui.setupUi(FormItensPendentesDetalhes)
    FormItensPendentesDetalhes.show()
    sys.exit(app.exec_())



"""
Pacote de Telemetria - Versão Simplificada

Este pacote implementa um sistema de telemetria básico com 8 eventos essenciais:
- Carga Inicial (3 eventos): ITG_INI_CARGA, ITG_CARGA_ENTIDADE, ITG_FIM_CARGA
- Integração Contínua (3 eventos): ITG_INI_INTEG, ITG_INTEG_ENTIDADE, ITG_FIM_INTEG
- Verificação de Integridade (2 eventos): ITG_INI_VERIF, ITG_FIM_VERIF
"""

from nsj_integracao_api_client.infra.telemetria.telemetria_service import TelemetriaService
from nsj_integracao_api_client.infra.telemetria.telemetria_client import TelemetriaClient
from nsj_integracao_api_client.infra.telemetria.telemetria_decorator import telemetria, telemetria_simples
from nsj_integracao_api_client.infra.telemetria.telemetria_metrics import TelemetriaMetrics, obter_metricas
from nsj_integracao_api_client.infra.telemetria.factory import TelemetriaFactory, inicializar_telemetria, obter_telemetria
from nsj_integracao_api_client.infra.telemetria.campos_fixos import obter_campos_fixos
from nsj_integracao_api_client.infra.telemetria.config import obter_config_ambiente, validar_config

__all__ = [
    # Serviços principais
    'TelemetriaService',
    'TelemetriaClient',
    'TelemetriaMetrics',

    # Decoradores
    'telemetria',
    'telemetria_simples',

    # Factory e inicialização
    'TelemetriaFactory',
    'inicializar_telemetria',
    'obter_telemetria',

    # Métricas
    'obter_metricas',

    # Campos fixos
    'obter_campos_fixos',

    # Configuração
    'obter_config_ambiente',
    'validar_config'
]



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'resources/form/verificacao_integridade.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_FormVerificacaoIntegridade(object):
    def setupUi(self, FormVerificacaoIntegridade):
        FormVerificacaoIntegridade.setObjectName("FormVerificacaoIntegridade")
        FormVerificacaoIntegridade.resize(360, 378)
        self.verticalLayout = QtWidgets.QVBoxLayout(FormVerificacaoIntegridade)
        self.verticalLayout.setObjectName("verticalLayout")
        self.labelDescricao = QtWidgets.QLabel(FormVerificacaoIntegridade)
        self.labelDescricao.setWordWrap(True)
        self.labelDescricao.setObjectName("labelDescricao")
        self.verticalLayout.addWidget(self.labelDescricao)
        self.groupBox = QtWidgets.QGroupBox(FormVerificacaoIntegridade)
        self.groupBox.setTitle("")
        self.groupBox.setFlat(False)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.checkTraceback = QtWidgets.QCheckBox(self.groupBox)
        self.checkTraceback.setObjectName("checkTraceback")
        self.verticalLayout_2.addWidget(self.checkTraceback)
        self.checkVerboso = QtWidgets.QCheckBox(self.groupBox)
        self.checkVerboso.setObjectName("checkVerboso")
        self.verticalLayout_2.addWidget(self.checkVerboso)
        self.checkDataNoLog = QtWidgets.QCheckBox(self.groupBox)
        self.checkDataNoLog.setObjectName("checkDataNoLog")
        self.verticalLayout_2.addWidget(self.checkDataNoLog)
        self.checkPararCasoDif = QtWidgets.QCheckBox(self.groupBox)
        self.checkPararCasoDif.setObjectName("checkPararCasoDif")
        self.verticalLayout_2.addWidget(self.checkPararCasoDif)
        self.checkDetalharDiff = QtWidgets.QCheckBox(self.groupBox)
        self.checkDetalharDiff.setObjectName("checkDetalharDiff")
        self.verticalLayout_2.addWidget(self.checkDetalharDiff)
        self.checkCorrigir = QtWidgets.QCheckBox(self.groupBox)
        self.checkCorrigir.setObjectName("checkCorrigir")
        self.verticalLayout_2.addWidget(self.checkCorrigir)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.labelTipoVerificacao = QtWidgets.QLabel(self.groupBox)
        self.labelTipoVerificacao.setObjectName("labelTipoVerificacao")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.labelTipoVerificacao)
        self.cmbTpVerificacao = QtWidgets.QComboBox(self.groupBox)
        self.cmbTpVerificacao.setObjectName("cmbTpVerificacao")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.cmbTpVerificacao)
        self.verticalLayout_2.addLayout(self.formLayout)
        self.verticalLayout.addWidget(self.groupBox)
        self.buttonExecutar = QtWidgets.QPushButton(FormVerificacaoIntegridade)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonExecutar.sizePolicy().hasHeightForWidth())
        self.buttonExecutar.setSizePolicy(sizePolicy)
        self.buttonExecutar.setObjectName("buttonExecutar")
        self.verticalLayout.addWidget(self.buttonExecutar)
        self.buttonExecutarRapida = QtWidgets.QPushButton(FormVerificacaoIntegridade)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonExecutarRapida.sizePolicy().hasHeightForWidth())
        self.buttonExecutarRapida.setSizePolicy(sizePolicy)
        self.buttonExecutarRapida.setObjectName("buttonExecutarRapida")
        self.verticalLayout.addWidget(self.buttonExecutarRapida)
        self.buttonVerLogs = QtWidgets.QPushButton(FormVerificacaoIntegridade)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonVerLogs.sizePolicy().hasHeightForWidth())
        self.buttonVerLogs.setSizePolicy(sizePolicy)
        self.buttonVerLogs.setEnabled(False)
        self.buttonVerLogs.setObjectName("buttonVerLogs")
        self.verticalLayout.addWidget(self.buttonVerLogs)

        self.retranslateUi(FormVerificacaoIntegridade)
        QtCore.QMetaObject.connectSlotsByName(FormVerificacaoIntegridade)

    def retranslateUi(self, FormVerificacaoIntegridade):
        _translate = QtCore.QCoreApplication.translate
        FormVerificacaoIntegridade.setWindowTitle(_translate("FormVerificacaoIntegridade", "Verificação de integridade"))
        self.labelDescricao.setText(_translate("FormVerificacaoIntegridade", "Compare os dados entre o ERP e a API e identifique inconsistências."))
        self.checkTraceback.setText(_translate("FormVerificacaoIntegridade", "Exibir trace."))
        self.checkVerboso.setText(_translate("FormVerificacaoIntegridade", "Log verboso."))
        self.checkDataNoLog.setText(_translate("FormVerificacaoIntegridade", "Imprimir data no log."))
        self.checkPararCasoDif.setText(_translate("FormVerificacaoIntegridade", "Parar caso encontre diferenças."))
        self.checkDetalharDiff.setText(_translate("FormVerificacaoIntegridade", "Detalhar diferenças."))
        self.checkCorrigir.setText(_translate("FormVerificacaoIntegridade", "Corrigir diferenças encontradas."))
        self.labelTipoVerificacao.setText(_translate("FormVerificacaoIntegridade", "Tipo verificação: "))
        self.buttonExecutar.setText(_translate("FormVerificacaoIntegridade", "▶   Executar Verificação Completa"))
        self.buttonExecutarRapida.setText(_translate("FormVerificacaoIntegridade", "⚡   Executar Verificação Rápida"))
        self.buttonVerLogs.setText(_translate("FormVerificacaoIntegridade", "📄   Visualizar Logs"))
